"""
================
Brainweb Phantom
================

Example of Brainweb phantom creation.

This examples show how to generate numerical phantoms based on the Brainweb
dataset.

"""

import matplotlib.pyplot as plt
import numpy as np

from mrtwin import brainweb_phantom

# %%
# Basic Usage
# ===========
# The Brainweb database consists of a set of 20 normal brains
# (ids equal to ``[4, 5, 6, 18, 20, 38, 41-54]``) at a nominal 0.5 mm isotropic
# resolution and a shape of ``(nz, ny, nx) = (362, 434, 362)``.
#
# A digital Brainweb phantom can be created as:

phantom2D = brainweb_phantom(ndim=2, subject=4, segtype="fuzzy")  # 2D phantom
phantom3D = brainweb_phantom(ndim=3, subject=4, segtype="fuzzy")  # 3D phantom

# %%
# The phantoms here created are sparse, i.e., they consists of a
# ``(nclasses, *spatial_shape)`` shaped ``np.ndarray`` representing the
# probabilistic maps of each tissue type (e.g., Gray Matter, White Matter, CSF)
# and a list of ``(nclasses,)`` dictionaries each containing the ``(M0, T1, T2, T2*, Chi)``
# values for each class:

example2D = np.concatenate((phantom2D[3], phantom2D[2], phantom2D[1]), axis=0)

example3Dax = np.concatenate(
    (phantom3D[3, 100], phantom3D[2, 100], phantom3D[1, 100]), axis=0
)
example3Dcor = np.concatenate(
    (phantom3D[3, ::-1, 100], phantom3D[2, ::-1, 100], phantom3D[1, ::-1, 100]), axis=0
)
example3Dsag = np.concatenate(
    (
        phantom3D[3, ::-1, :, 100],
        phantom3D[2, ::-1, :, 100],
        phantom3D[1, ::-1, :, 100],
    ),
    axis=0,
)
example3D = np.concatenate((example3Dax, example3Dcor, example3Dsag), axis=1)

fig1, ax1 = plt.subplots(1, 2)
ax1[0].imshow(example2D, cmap="gray"), ax1[0].axis("off"), ax1[0].set_title(
    "2D phantom"
)
ax1[1].imshow(example3D, cmap="gray"), ax1[1].axis("off"), ax1[1].set_title(
    "3D phantom"
)
plt.show()

# %%
# The ``(M0, T1, T2, T2*, Chi)`` properties
# can be direcly accessed as:

_, _ = print("M0:", end="\t"), print(phantom2D.M0)  # same for phantom3D
_, _ = print("T1 (ms):", end="\t"), print(phantom2D.T1)
_, _ = print("T2 (ms):", end="\t"), print(phantom2D.T2)
_, _ = print("T2* (ms):", end="\t"), print(phantom2D.T2s)
_, _ = print("Chi:", end="\t"), print(phantom2D.Chi)

# %%
# If required, the ``properties`` dictionary
# can be directly accessed as:

print(phantom2D.properties)

# %%
# e.g., to be passed as ``**kwargs`` to a simulator routine.
#
# Notice that segmentation can be accessed directly (in read-only mode)
# via square bracked indexing, similarly to numpy arrays.
#
# A basic summary of the properties can be accessed
# via the ``__repr__`` attribute (i.e., enabling pretty printing):

print(phantom2D)
print(phantom3D)

# We also provide a crisp segmentation, which has a lower memory footprint
# at cost of a coarser approximation (i.e., a piecewise-constant tissue model)-
#
# This can be obtained starting from the `"fuzzy"` phantom as:

phantom2D = phantom2D.as_crisp()
phantom3D = phantom3D.as_crisp()

# Print summary

print(phantom2D)
print(phantom3D)

# Display spatial segmentations

example2D = phantom2D

example3Dax = np.concatenate((phantom3D[100], phantom3D[100], phantom3D[100]), axis=0)
example3Dcor = np.concatenate(
    (phantom3D[::-1, 100], phantom3D[::-1, 100], phantom3D[::-1, 100]), axis=0
)
example3Dsag = np.concatenate(
    (phantom3D[::-1, :, 100], phantom3D[::-1, :, 100], phantom3D[::-1, :, 100]), axis=0
)
example3D = np.concatenate((example3Dax, example3Dcor, example3Dsag), axis=1)

fig2, ax2 = plt.subplots(1, 2)
ax2[0].imshow(example2D, cmap="turbo"), ax2[0].axis("off"), ax2[0].set_title(
    "2D phantom"
)
ax2[1].imshow(example3D, cmap="turbo"), ax2[1].axis("off"), ax2[1].set_title(
    "3D phantom"
)
plt.show()

# %%
# Crisp phantom can be also directly generated as:

phantom2D = brainweb_phantom(ndim=2, subject=4, segtype="crisp")  # 2D phantom
phantom3D = brainweb_phantom(ndim=3, subject=4, segtype="crisp")  # 3D phantom

# N.B. ``segtype`` can be omitted as the default is ``"crisp"``.

phantom2D = brainweb_phantom(ndim=2, subject=4)  # single-slice 2D phantom
phantom3D = brainweb_phantom(ndim=3, subject=4)  # 3D phantom

# %%
# Finally, we can obtain a "dense" phantom,
# i.e., an object without segmentation whose
# ``(M0, T1, T2, T2*, Chi)`` properties are stored
# as parametric maps rather than the individual values
# of each tissue class.
#
# This can be obtain (both from ``"fuzzy"`` and ``"crisp"`` models) as:

phantom2D = phantom2D.as_numeric()

# Print summary

print(phantom2D)

# Display parameter maps

fig3, ax3 = plt.subplots(1, 5)

im0 = ax3[0].imshow(phantom2D.M0, cmap="gray")
ax3[0].axis("off"), ax3[0].set_title("M0 [a.u.]")
fig3.colorbar(im0, ax=ax3[0], fraction=0.046, pad=0.04)

im1 = ax3[1].imshow(phantom2D.T1, cmap="magma")
ax3[1].axis("off"), ax3[1].set_title("T1 [ms]")
fig3.colorbar(im1, ax=ax3[1], fraction=0.046, pad=0.04)

im2 = ax3[2].imshow(phantom2D.T2, cmap="viridis", vmax=150)
ax3[2].axis("off"), ax3[2].set_title("T2 [ms]")
fig3.colorbar(im2, ax=ax3[2], fraction=0.046, pad=0.04)

im3 = ax3[3].imshow(phantom2D.T2s, cmap="viridis", vmax=150)
ax3[3].axis("off"), ax3[3].set_title("T2* [ms]")
fig3.colorbar(im3, ax=ax3[3], fraction=0.046, pad=0.04)

im4 = ax3[4].imshow(phantom2D.Chi, cmap="gray")
ax3[4].axis("off"), ax3[4].set_title("Chi")
fig3.colorbar(im4, ax=ax3[4], fraction=0.046, pad=0.04)

plt.tight_layout()
plt.show()

# %%
# Dense phantom can be also directly generated as:

phantom2D = brainweb_phantom(ndim=2, subject=4, segtype=False)  # 2D phantom

# Print summary

print(phantom2D)

# %%
# Hereafter, without loss of generality, we will use 2D phantoms.

# %% Setting spatial properties
#
# By default, Brainweb phantoms are interpolated
# to 1.085 mm isotropic resolution with a 200 isotropic matrix (217 mm iso FOV).
#
# These can be adjusted using ``shape`` and ``output_res`` model:
#
# 1. ``shape`` will control the matrix size without affecting the resolution.
# 2. ``output_res`` will adjust the spatial resolution keeping the same (i.e., 200 iso) matrix
#

phantom2D = brainweb_phantom(ndim=2, subject=4)
phantom2D_mtx = brainweb_phantom(
    ndim=2, subject=4, shape=256
)  # can also be shape=(ny, nx)
phantom2D_res = brainweb_phantom(
    ndim=2, subject=4, output_res=2.0
)  # can also be output_res=(dy, dx)
phantom2D_mtx_res = brainweb_phantom(ndim=2, subject=4, shape=256, output_res=2.0)

print(phantom2D)
print(phantom2D_mtx)
print(phantom2D_res)
print(phantom2D_mtx_res)

fig4, ax4 = plt.subplots(2, 2)
ax4[0, 0].imshow(phantom2D, cmap="turbo"), ax4[0, 0].axis("off"), ax4[0, 0].set_title(
    "shape 200, res=1.085mm, fov=217mm"
)
ax4[0, 1].imshow(phantom2D_mtx, cmap="turbo"), ax4[0, 1].axis("off"), ax4[
    0, 1
].set_title("shape 256, res=1.085mm, fov=277mm")
ax4[1, 0].imshow(phantom2D_res, cmap="turbo"), ax4[1, 0].axis("off"), ax4[
    1, 0
].set_title("shape 200, res=2.0mm, fov=400mm")
ax4[1, 1].imshow(phantom2D_mtx_res, cmap="turbo"), ax4[1, 1].axis("off"), ax4[
    1, 1
].set_title("shape 256, res=2.0mm, fov=512mm")
plt.show()


# %% Setting field strength
#
# The physical parameter of each tissue class are calculated by
# default for a field strength of 1.5 T.
#
# This can be changed via the ``B0`` argument:

# B0 strengths
B0 = [0.55, 1.5, 3.0, 7.0, 11.7, 13.3]  # field strengths in [T]

# Generate phantoms with different field strengths
phantomB0 = [
    brainweb_phantom(ndim=2, subject=4, B0=strength, segtype=False) for strength in B0
]

# Display
T1 = np.concatenate([phantom.T1 for phantom in phantomB0], axis=1)
T2 = np.concatenate([phantom.T2 for phantom in phantomB0], axis=1)
T2s = np.concatenate([phantom.T2s for phantom in phantomB0], axis=1)

fig5, ax5 = plt.subplots(3, 1)

im1 = ax5[0].imshow(T1, cmap="magma", vmax=5000)
ax5[0].axis("off"), ax5[0].set_title("T1 [ms]")
fig5.colorbar(im1, ax=ax5[0], fraction=0.046, pad=0.04)

im2 = ax5[1].imshow(T2, cmap="viridis", vmax=150)
ax5[1].axis("off"), ax5[1].set_title("T2 [ms]")
fig5.colorbar(im2, ax=ax5[1], fraction=0.046, pad=0.04)

im3 = ax5[2].imshow(T2s, cmap="viridis", vmax=100)
ax5[2].axis("off"), ax5[2].set_title("T2* [ms]")
fig5.colorbar(im3, ax=ax5[2], fraction=0.046, pad=0.04)

plt.tight_layout()
plt.show()


# %% Advanced tissue models
#
# In addition to single pool model, we provide 3 multi-pool models:
#
# 1. ``"mw-model"``: a two-pool model where free water is divided in two compartments,
#    i.e., `intra-/extra-cellular water` (long T1 / T2) and `myelin water` (short T1 / T2).
#    The model include chemical exchange between the two pools.
#    Parameters are ``(MWF, T1, T2, k, chemshift)``.
# 2. ``"mt-model"``: a two-pool model consisting in `free water` and `bound water`.
#    Free water includes both intra-/extra-cellular and myelin water (as in the single-pool model),
#    while bound water corresponds to a macromolecular pool with the same T1 as the free water
#    and no T2 (i.e., no transverse magnetiztion).
#    The model include magnetization transfer between the two pools.
#    Parameters are ``(MVF, T1, T2, k)``.
# 3. ``"mwmt-model"``: a three-pool model consisting in `intra-/extra-cellular water`, `myelin water` and `bound water`.
#    The model include chemical exchange between the two free water pools and magnetizion transfer between
#    myelin water and bound water.
#    Parameters are ``(MWF, MVF, T1, T2, k)``.
#
# Here we will display the latter, as it represents the most general case.

# model="single-pool" is the default, while "mw-model" and "mt-model" corresponds to cases 1. and 2.
phantom_multi = brainweb_phantom(ndim=2, subject=4, model="mwmt-model", segtype=False)

# %%
#
# MWF corresponds to the myelin water fraction, while MVF to the bound water fraction.
# We assume that intra-extracellular water fraction ``= 1 - (MWF + MVF)``:

MWF = phantom_multi.MWF
MVF = phantom_multi.MVF
IEWF = (1 - (MWF + MVF)) * (MWF > 0)
weight = np.concatenate((IEWF, MWF, MVF), axis=1)

plt.figure()
plt.imshow(weight, vmin=0, vmax=1, cmap="hot"), plt.axis("off"), plt.title(
    "pool fractions"
), plt.colorbar()
plt.show()

# %%
#
# T1 and T2 for the two free water pools are stacked along the first axis,
# with ``n=0`` being the intra-/extra-cellular water (long T1 / T2) and
# ``n=1`` being the myelin water (short T1 / T2):

T1 = np.concatenate((phantom_multi.T1[0], phantom_multi.T1[1]), axis=1)
T2 = np.concatenate((phantom_multi.T2[0], phantom_multi.T2[1]), axis=1)

fig6, ax6 = plt.subplots(2, 1)

im1 = ax6[0].imshow(T1, cmap="magma", vmax=1500)
ax6[0].axis("off"), ax6[0].set_title("T1 [ms]")
fig6.colorbar(im1, ax=ax6[0], fraction=0.046, pad=0.04)

im2 = ax6[1].imshow(T2, cmap="viridis", vmax=150)
ax6[1].axis("off"), ax6[1].set_title("T2 [ms]")
fig6.colorbar(im2, ax=ax6[1], fraction=0.046, pad=0.04)


# %%
#
# k represent the non-directional exchange rates in [Hz],
# with ``n=0`` being the chemical exchange rate between the two free water pools
# and ``n=1`` being magnetization transfer rate between the myelin and bound water:

k = np.concatenate((phantom_multi.k[0], phantom_multi.k[1]), axis=1)

plt.figure()
plt.imshow(k, cmap="hot"), plt.axis("off"), plt.title(
    "exchange rate [Hz]"
), plt.colorbar()
plt.show()

# %%
#
# Similarly to the single pool model, ``mrtwin`` supports ``"fuzzy"`` and ``"crisp"``
# segmentations.
#
# Caching mechanism
# =================
#
# To reduce loading times, `mrtwin` implements a caching mechanism.
#
# If ``cache`` argument is set to ``True`` (default behaviour), each phantom
# segmentation (identified by the number of spatial dimensions,
# tissue model, segmentation type, matrix shape and resolution)
# is saved on the disk in ``npy`` format.
#
# The path is selected according to the following hierachy (inspired by ``brainweb-dl``):
#
# 1. User-specific argument (``cache_dir``)
# 2. ``MRTWIN_DIR`` environment variable
# 3. ``~/.cache/mrtwin`` folder
#
# On top of that, the underlying Brainweb fuzzy segmentations at the original
# resolution are stored (via ``brainweb-dl``) with the following hierachy:
#
# 1. User-specific argument (``brainweb_dir``)
# 2. ``BRAINWEB_DIR`` environment variable
# 3. ``~/.cache/brainweb`` folder
#
# The cached files can be forcibly overwritten by setting the ``force`` argument
# to ``True`` (default: ``False``).
#
# N.B.: currently, ``force=True`` will both force re-downloading brainweb segmentation
# and the subsequent interpolations and resizing required to obtain the
# desired shape and resolution.
#
# Disabling SSL verification (not recommended)
# ============================================
#
# If you encounter some issue in downloading, SSL verification
# can be disabled by setting ``verify`` to ``False`` (default: ``True``).
# It is advised however to solve the problem on your machine side
# (updating the certificate).



"""Single-pool Open Science CBS phantom builder class."""

__all__ = ["NumericOSFPhantom"]

from typing import Sequence


from .. import _classes

from .._build import PhantomMixin
from .._utils import CacheDirType

from ._base import OSFPhantom

fudge_factor = 1.5  # MRF has lower value wrt SE


class NumericOSFPhantom(OSFPhantom, PhantomMixin):
    """Numeric OSF phantom builder."""

    def __init__(
        self,
        ndim: int,
        subject: int,
        shape: int | Sequence[int] = None,
        output_res: float | Sequence[float] = None,
        B0: float = 3.0,
        cache: bool = True,
        cache_dir: CacheDirType = None,
        osf_dir: CacheDirType = None,
        force: bool = False,
        verify: bool = True,
    ):

        # initialize segmentation
        super().__init__(
            ndim,
            subject,
            shape,
            output_res,
            cache,
            cache_dir,
            osf_dir,
            force,
            verify,
        )

        # initialize model
        self.get_model(B0)

    def get_model(self, B0: float):
        """Initialize model.

        Parameters
        ----------
        B0, float
            Static field strength in [T].

        """
        self._properties = {}
        self.properties["M0"] = self.maps[0]
        self.properties["T1"] = _classes.extrapolate_t1(self.maps[1], 3.0, B0)
        self._properties["T2"] = fudge_factor * self.maps[2]
        self._properties["T2s"] = _classes.extrapolate_t2star(
            self.maps[3],
            self._properties["T2"],
            3.0,
            B0,
        )
        self.properties["Chi"] = self.maps[4]

    @property
    def M0(self):  # noqa
        return self._properties["M0"]

    @property
    def T1(self):  # noqa
        return self._properties["T1"]

    @property
    def T2(self):  # noqa
        return self._properties["T2"]

    @property
    def T2s(self):  # noqa
        return self._properties["T2s"]

    @property
    def Chi(self):  # noqa
        return self._properties["Chi"]

    @property
    def properties(self):  # noqa
        return self._properties



"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys
import coverage

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../.."))  # Source code dir relative to this file

# -- Project information -----------------------------------------------------

project = "mrtwin"
copyright = "2024, MRTwin Contributors"
author = "MRTwin Contributors"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_copybutton",
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_gallery.gen_gallery",
    "sphinx_add_colab_link",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# generate autosummary even if no references
autosummary_generate = True
# autosummary_imported_members = True
autodoc_inherit_docstrings = True
autodoc_member_order = "bysource"
autodoc_typehints = "description"

napoleon_include_private_with_doc = True
napolon_numpy_docstring = True
napoleon_use_admonition_for_references = True


pygments_style = "sphinx"
highlight_language = "python"

# -- Options for Sphinx Gallery ----------------------------------------------

sphinx_gallery_conf = {
    "doc_module": "mrtwin",
    "backreferences_dir": "generated/gallery_backreferences",
    "reference_url": {"mrtwin": None},
    "examples_dirs": ["../examples/"],
    "gallery_dirs": ["generated/autoexamples"],
    "filename_pattern": "/example_",
    "ignore_pattern": r"(__init__|conftest|utils).py",
    "nested_sections": True,
    "binder": {
        "org": "infn-mri",
        "repo": "mrtwin",
        "branch": "gh-pages",
        "binderhub_url": "https://mybinder.org",
        "dependencies": [
            "./binder/apt.txt",
            "./binder/environment.yml",
        ],
        "notebooks_dir": "examples",
        "use_jupyter_lab": True,
    },
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "sphinx_book_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]
html_theme_options = {
    "repository_url": "https://github.com/INFN-MRI/mrtwin",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "use_download_button": True,
    "home_page_in_toc": True,
}

# html_logo = "_static/logos/mri-nufft.png"
# html_favicon = "_static/logos/mri-nufft-icon.png"
html_title = "MRTwin Documentation"



"""
=========================
Transmit Field Simulation
=========================

Example of transmit field (i.e., B1+) map generation.

This examples show how to generate coil sensitivity maps
for a single or multi-channel transmit array.
"""

import matplotlib.pyplot as plt
import numpy as np

from mrtwin import b1field

plt.rcParams["image.cmap"] = "hot"


# %%
# Basic Usage
# ===========
#
# Two- and three-dimensional transmit field maps
# can be generated providing ``(ny, nx)`` and
# ``(nz, ny, nx)`` shaped tuple as a ``shape`` argument
# to `b1field` routine, respectively:

b1map2D = b1field(shape=(200, 200))  # (200, 200) matrix
print(b1map2D.shape)

b1map3D = b1field(shape=(128, 128, 128))  # (128, 128, 128) matrix
print(b1map3D.shape)

plt.figure()
plt.imshow(b1map2D), plt.axis("off"), plt.colorbar(), plt.title(
    "transmit field (in units of relative flip angle)"
)
plt.show()

# %% Setting field inhomgeneity
#
# The transmit field maps are generated
# with a relative flip angle variation
# between 0.8 and 1.2 (e.g., 3T systems).
#
# This can be changed via the ``b1range`` argument:

b1map2D = b1field(shape=(200, 200), b1range=(0.5, 2.0))

plt.figure()
plt.imshow(b1map2D), plt.axis("off"), plt.colorbar(), plt.title(
    "transmit field with b1range between 0.5 and 2.0"
)
plt.show()

# %%
#
# This can be used e.g., to simulate field maps for higher fields (e.g., 7T systems).
#
# Optionally, we can provide a mask of the object to exclude the background
# when calculating the field rescaling:

from mrtwin import shepplogan_phantom

mask = shepplogan_phantom(ndim=2, shape=(200, 200), segtype=False).M0 != 0.0
b1map2D = b1field(shape=(200, 200), mask=mask)

plt.figure()
plt.imshow(b1map2D), plt.axis("off"), plt.colorbar(), plt.title("masked transmit field")
plt.show()

# %%
# Advanced Options
# ================
#
# The sensitivity maps can be altered by modifying several parameters:
#
# 1. ``coil_width``: width of the coil (with respect to FOV).
# 2. ``shift``: displacement of the center (in units of voxels).
# 3. ``dphi``: bulk rotation of the coil (in [deg]).
# 4. ``ncoils``: number of transmit channels in the transmit array.
# 5. ``nrings``: number of rings for a cylindrical hardware setup.
#
# Without loss of generality, we show examples for 2D sensitivities:

# %%
# Coil width
# ----------
widths = [0.5, 1.0, 1.5, 2.0]
b1map2D = [
    b1field(shape=(200, 200), coil_width=w) for w in widths
]  # only show first channel

display = np.concatenate(b1map2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.colorbar(), plt.title(
    "coil width from 0.5 to 2.0 times fov"
)
plt.show()

# %%
# Center shift
# ------------
dx = [-20, -10, 0, 10, 20]
b1map2D = [
    b1field(shape=(200, 200), shift=(0, x), coil_width=0.5) for x in dx
]  # for 3D, it would be shift=(dz, dy, dx)

display = np.concatenate(b1map2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.colorbar(), plt.title(
    "x-displacement from -20 to 20 times voxels"
)
plt.show()

# %%
# Rotation
# --------
phi = [-20, -10, 0, 10, 20]
b1map2D = [b1field(shape=(200, 200), dphi=angle, coil_width=0.5) for angle in phi]

display = np.concatenate(b1map2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.colorbar(), plt.title(
    "coil rotation from -20 to 20 degrees (first channel)"
)
plt.show()

# %%
# Number of rings
# ---------------
ncoils = [1, 2, 4, 8, 16, 32]
b1map2D = [b1field(shape=(200, 200), ncoils=n, coil_width=0.5) for n in ncoils]

display = np.concatenate(b1map2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.colorbar(), plt.title(
    "number of channels from 1 to 16"
)
plt.show()


# %%
# Number of rings
# ---------------
nrings = [2, 4, 6, 8, 10]
b1map2D = [b1field(shape=(200, 200), nrings=n, coil_width=0.5) for n in nrings]

display = np.concatenate(b1map2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.colorbar(), plt.title(
    "number of rings from 2 to 10 (first channel)"
)
plt.show()


# %%
# Multiple Transmit Modes
# =======================
#
# By default, the sensitivities from each transmit channel are combined
# in quadrature mode.
#
# With ``mrtwin``, multiple orthogonal modes can be simulated by ``nmodes`` argument.
# For example, `CP` mode and `gradient` modes (e.g., for static pTx)  can be obtained as:

b1map = b1field((200, 200), nmodes=2)  # b1map[0] is CP, b1map[1] is gradient mode.

# %%
#
# In this case, b1map will be a ``(nmodes, *shape)`` ``np.ndarray``, the different
# modes being stacked along the first axis:

fig, ax = plt.subplots(2, 2)
im0 = ax[0, 0].imshow(np.abs(b1map[0]))
ax[0, 0].axis("off"), ax[0, 0].set_title("CP mode (magn)")
fig.colorbar(im0, ax=ax[0, 0], fraction=0.046, pad=0.04)

im1 = ax[0, 1].imshow(np.angle(b1map[0]))
ax[0, 1].axis("off"), ax[0, 1].set_title("CP mode (phase)")
fig.colorbar(im1, ax=ax[0, 1], fraction=0.046, pad=0.04)

im2 = ax[1, 0].imshow(np.abs(b1map[1]))
ax[1, 0].axis("off"), ax[1, 0].set_title("gradient mode (magn)")
fig.colorbar(im2, ax=ax[1, 0], fraction=0.046, pad=0.04)

im3 = ax[1, 1].imshow(np.angle(b1map[1]))
ax[1, 1].axis("off"), ax[1, 1].set_title("gradient mode (phase)")
fig.colorbar(im3, ax=ax[1, 1], fraction=0.046, pad=0.04)
plt.show()

# %%
# Caching mechanism
# =================
#
# To reduce loading times, ``mrtwin`` implements a caching mechanism.
#
# If ``cache`` argument is set to ``True`` (default behaviour for ``ndim=3``), each transmit field map
# segmentation (identified by the number of channels, number of modes,
# matrix size, shift, rotation angle, number of rings, b1 range and masking flag)
# is saved on the disk in ``npy`` format.
#
# The path is selected according to the following hierachy (inspired by ``brainweb-dl``):
#
# 1. User-specific argument (``cache_dir``)
# 2. ``MRTWIN_DIR`` environment variable
# 3. ``~/.cache/mrtwin`` folder
#



r"""
=======================
Static Field Simulation
=======================

Example of static field generation.

This examples show how to generate a static field (i.e., B0)
map starting from the object susceptibility distribution.


The static field is written as the following convolution (performed as a pointwise multiplication in the Fourier space):

.. math::
    
    \mathbf{B}(\mathbf{r}) = \mathbf{B_0} \cdot \mathcal{F}^{-1} \left[ \chi(\mathbf{k}) \cdot D(\mathbf{k}) \right]
    
where :math:`\chi(\mathbf{k})` is the Fourier transform of susceptibility spatial distribution and
:math:`D(\mathbf{k})` is the dipole kernel defined by:
    
.. math ::

    D(\mathbf{r}) = \left(\frac{1}{3} - \frac{k^2_z}{\lvert \textbf{k} \rvert}\right)

"""

import matplotlib.pyplot as plt
import numpy as np

from mrtwin import shepplogan_phantom, b0field

# %%
# Generate susceptibility distribution
# ====================================
#
# First, we generate a 3D Shepp-Logan phantom:

phantom = shepplogan_phantom(ndim=3, shape=128, segtype=False)

# Get susceptibility
chi = phantom.Chi

plt.figure()
plt.imshow(chi[64], cmap="gray"), plt.axis("off"), plt.colorbar(), plt.title(
    "magnetic susceptibility"
)
plt.show()

# %%
# Basic Usage
# ===========
#
# Starting from the susceptibility distribution, we can generate the
# corresponding static field perturbation as:

b0map = b0field(chi)

plt.figure()
plt.imshow(b0map[64], cmap="turbo", vmin=-300, vmax=300), plt.axis(
    "off"
), plt.colorbar(), plt.title("B0 map [Hz]")
plt.show()

# %%
# Optionally, we can provide a mask of the object to exclude the background:

mask = phantom.M0 != 0.0
b0map = b0field(chi, mask=mask)

plt.figure()
plt.imshow(b0map[64], cmap="turbo", vmin=-300, vmax=300), plt.axis(
    "off"
), plt.colorbar(), plt.title("masked B0 map [Hz]")
plt.show()

# %% Setting field strength
#
# The static field map calculated by
# default for a field strength of 1.5 T.
#
# This can be changed via the ``B0`` argument:

# B0 strengths
B0 = [0.55, 1.5, 3.0, 7.0, 11.7, 13.3]  # field strengths in [T]

# Generate phantoms with different field strengths
b0maps = [b0field(chi, B0=strength) for strength in B0]

# Display
b0map = np.concatenate([b0map[64] for b0map in b0maps], axis=1)

plt.figure()
plt.imshow(b0map, cmap="turbo", vmin=-300, vmax=300), plt.axis(
    "off"
), plt.colorbar(), plt.title("B0 map [Hz]")
plt.show()

# %%
# As an alternative, we can force a specific
# B0 offset range using the ``b0range`` argument as:

b0map = b0field(chi, b0range=(-500, 500))

plt.figure()
plt.imshow(b0map[64], cmap="turbo", vmin=-300, vmax=300), plt.axis(
    "off"
), plt.colorbar(), plt.title("B0 map [Hz]")
plt.show()



"""Test Shepp-Logan phantom generation."""

import os
import tempfile


import pytest


import numpy.testing as npt


from mrtwin import shepplogan_phantom


@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("shape", [128, (128, 128), (128, 128, 128)])
@pytest.mark.parametrize("model", ["single-pool", "mw-model", "mt-model", "mwmt-model"])
@pytest.mark.parametrize("segtype", ["crisp", False])
def test_shepplogan_phantom(ndim, shape, model, segtype):
    """
    Test the shepplogan_phantom function with various parameter combinations.
    """
    if ndim == 2 and shape == (128, 128, 128):
        pytest.skip("2D phantom not compatible with len(shape)==3")
    if ndim == 3 and shape == (128, 128):
        pytest.skip("3D phantom not compatible with len(shape)==2")

    phantom = shepplogan_phantom(ndim=ndim, shape=shape, model=model, segtype=segtype)

    # Validate the output is not None
    assert phantom is not None, "Phantom output should not be None."

    # Validate the shape of the output if a shape is provided
    if shape is not None:
        expected_shape = tuple([shape] * ndim) if isinstance(shape, int) else shape
        actual_shape = phantom.shape[-ndim:] if segtype else phantom.T1.shape[-ndim:]
        npt.assert_allclose(actual_shape, expected_shape)



"""
Wrapper around brainweb-dl to extract tissue segmentation.

N.B. I did not directly used get_mri to allow for
SSL verification disabling (not advised, use with care).
"""

__all__ = ["get_brainweb_segmentation"]

import gzip
import io
import logging
import os
import requests
import warnings

from typing import Sequence

import numpy as np

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import brainweb_dl

from pathlib import Path

from numpy.typing import DTypeLike

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from tqdm.auto import tqdm

from brainweb_dl._brainweb import (
    BASE_URL,
    SUB_ID,
    STD_RES_SHAPE,
    load_array,
)

from .. import _prescription

from .._utils import ssl_verification, CacheDirType


def _request_get_brainweb(
    download_command: str,
    path: os.PathLike,
    force: bool = False,
    dtype: DTypeLike = np.float32,
    shape: tuple = STD_RES_SHAPE,
) -> np.ndarray:
    """Request to download brainweb dataset.

    Parameters
    ----------
    download_command : str
        Formatted request code to download a volume from brainweb.
    path : os.PathLike
        Path to save the downloaded file.
    force : bool, optional
        Force download even if the file already exists.
        The default is False.
    dtype : DTypeLike, optional
        Data type of the downloaded file.
        The default is np.float32.
    shape : tuple, optional
        Shape of the downloaded file.
        The default is (182, 217, 181)

    Returns
    -------
    np.ndarray
        Downloaded file.

    Raises
    ------
    Exception
        If the download fails.
    """
    path = Path(path)
    # don't download if it cached.
    if path.exists() and not force:
        return load_array(path)
    d = requests.get(
        _get_url(BASE_URL, download_command),
        stream=True,
        headers={"Accept-Encoding": "identity", "Content-Encoding": "gzip"},
    )

    # download
    with (
        io.BytesIO() as buffer,
        tqdm(
            total=float(d.headers.get("Content-length", 0)),
            desc=f"Downloading {download_command}",
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            leave=False,
            position=2,
        ) as pbar,
    ):
        for chunk in d.iter_content(chunk_size=1024):
            buffer.write(chunk)
            pbar.update(len(chunk))
        data = np.frombuffer(gzip.decompress(buffer.getvalue()), dtype=dtype)
    if data.size != np.prod(shape):
        raise ValueError(f"Mismatch between data size and shape {data.size} != {shape}")
    data = abs(data).reshape(shape)
    return data


def _get_url(BASE_URL, download_command):
    URL = (
        BASE_URL
        + "?"
        + "&".join(
            [
                f"{k}={v}"
                for k, v in {
                    "do_download_alias": download_command,
                    "format_value": "raw_short",
                    "zip_value": "gnuzip",
                    "who_name": "",
                    "who_institution": "",
                    "who_email": "",
                    "download_for_real": "%5BStart+download%21%5D",
                }.items()
                if v
            ]
        )
    )
    return URL


# Monkey patch
brainweb_dl._brainweb._request_get_brainweb = _request_get_brainweb

# Actual functions
logger = logging.getLogger("brainweb_dl")


def get_brainweb_segmentation(
    ndim: int,
    subject: int,
    shape: int | Sequence[int] | None = None,
    output_res: float | Sequence[float] | None = None,
    brainweb_dir: CacheDirType = None,
    force: bool = False,
    verify: bool = True,
):
    """
    Get fuzzy BrainWeb tissue segmentation.

    Parameters
    ----------
    ndim : int
        Number of spatial dimensions. If ndim == 2, use a single slice
        (central axial slice).
    subject : int
        Subject id to download.
    shape: int | Sequence[int] | None, optional
        Shape of the output data, the data will be interpolated to the given shape.
        If int, assume isotropic matrix. The default is None (original shape).
    output_res: float | Sequence[float] | None, optional
        Resolution of the output data, the data will be rescale to the given resolution.
        If scalar, assume isotropic resolution. The default is None (1mm iso).
    brainweb_dir : CacheDirType, optional
        Brainweb_directory to download the data.
        The default is None (~/.cache/brainweb).
    force : bool, optional
        Force download even if the file already exists.
        The default is False.
    verify : bool, optional
        Enable SSL verification.
        DO NOT DISABLE (i.e., verify=False)IN PRODUCTION.
        The default is True.

    Returns
    -------
    np.ndarray.
        Brainweb segmentation.

    """
    assert ndim == 2 or ndim == 3, ValueError(
        f"Number of spatial dimensions (={ndim}) must be either 2 or 3."
    )
    assert subject in SUB_ID, ValueError(
        f"subject (={subject}) must be one of {SUB_ID}"
    )
    logger.debug(f"Get MRI data for subject {subject:02d}")

    # default params
    if shape is not None and np.isscalar(shape):
        shape = shape * np.ones(ndim, dtype=int)
    if shape is not None:
        assert len(shape) == ndim, ValueError(
            "If shape is not None, it must be either a scalar or a ndim-length sequence."
        )
    if output_res is not None and np.isscalar(output_res):
        output_res = output_res * np.ones(ndim, dtype=int)
    if output_res is not None:
        assert len(output_res) == ndim, ValueError(
            "If output_res is not None, it must be either or a scalar a ndim-length sequence."
        )

    # original resolution (0.5 mm iso)
    orig_res = 0.5 * np.ones(ndim)

    # get data
    if verify is False:
        with ssl_verification(verify=verify):
            data = brainweb_dl.get_mri(
                subject,
                "fuzzy",
                brainweb_dir=brainweb_dir,
                force=force,
            )
    else:
        data = brainweb_dl.get_mri(
            subject,
            "fuzzy",
            brainweb_dir=brainweb_dir,
            force=force,
        )

    # put tissue classes as leading axis
    data = data.transpose(-1, 0, 1, 2)
    data = np.flip(data, axis=-2)

    # select single slice
    if ndim == 2:
        center = int(data.shape[-3] // 2)
        data = data[:, center, :, :]

    # make sure it is contiguous
    data = np.ascontiguousarray(data)

    if shape is None and output_res is None:
        # normalize probability
        with np.errstate(divide="ignore", invalid="ignore"):
            data = data / data.sum(axis=0)
        data = np.nan_to_num(data, posinf=0.0, neginf=0.0)
        return data.astype(np.float32)
    elif output_res is None:
        output_res = 2 * orig_res  # 1 mm iso

    # set prescription
    data = _prescription.set_prescription(
        data, orig_res, data.shape[-ndim:], output_res, shape
    )

    # normalize probability
    with np.errstate(divide="ignore", invalid="ignore"):
        data = data / data.sum(axis=0)
    data = np.nan_to_num(data, posinf=0.0, neginf=0.0)

    return data.astype(np.float32)



"""Single-pool BrainWeb phantom builder class."""

__all__ = ["FuzzyBrainwebPhantom", "CrispBrainwebPhantom", "NumericBrainwebPhantom"]

import os
from typing import Sequence

import numpy as np

from .. import _classes

from .._build import FuzzyPhantomMixin, CrispPhantomMixin, _fuzzy_to_crisp
from .._utils import CacheDirType

from ._base import BrainwebPhantom


class FuzzyBrainwebPhantom(BrainwebPhantom, FuzzyPhantomMixin):
    """Fuzzy BrainWeb phantom builder."""

    def __init__(
        self,
        ndim: int,
        subject: int,
        shape: int | Sequence[int] = None,
        output_res: float | Sequence[float] = None,
        B0: float = 1.5,
        cache: bool = True,
        cache_dir: CacheDirType = None,
        brainweb_dir: CacheDirType = None,
        force: bool = False,
        verify: bool = True,
    ):

        # initialize segmentation
        super().__init__(
            ndim,
            subject,
            shape,
            output_res,
            cache,
            cache_dir,
            brainweb_dir,
            force,
            verify,
        )

        # initialize model
        self.get_model(B0)

    def get_model(self, B0: float):
        """Initialize model.

        Parameters
        ----------
        B0, float
            Static field strength in [T].

        """
        model_dict = _classes.tissue_map("single-pool")
        self._label = []
        self._properties = {"M0": [], "T1": [], "T2": [], "T2s": [], "Chi": []}
        for tissue in model_dict:
            self._label.append(tissue["Label"])
            self._properties["M0"].append(tissue["M0"])
            self._properties["T1"].append(_classes.get_t1(tissue, B0, 1.5))
            self._properties["T2"].append(tissue["T2"])
            self._properties["T2s"].append(_classes.get_t2star(tissue, B0, 1.5))
            self._properties["Chi"].append(tissue["Chi"])

        # cast to array
        self._label = np.asarray(self._label, dtype=int)
        self._properties["M0"] = np.asarray(self._properties["M0"], dtype=np.float32)
        self._properties["T1"] = np.asarray(self._properties["T1"], dtype=np.float32)
        self._properties["T2"] = np.asarray(self._properties["T2"], dtype=np.float32)
        self._properties["T2s"] = np.asarray(self._properties["T2s"], dtype=np.float32)
        self._properties["Chi"] = np.asarray(self._properties["Chi"], dtype=np.float32)

    @property
    def M0(self):  # noqa
        return self._properties["M0"]

    @property
    def T1(self):  # noqa
        return self._properties["T1"]

    @property
    def T2(self):  # noqa
        return self._properties["T2"]

    @property
    def T2s(self):  # noqa
        return self._properties["T2s"]

    @property
    def Chi(self):  # noqa
        return self._properties["Chi"]

    @property
    def properties(self):  # noqa
        return self._properties


class CrispBrainwebPhantom(FuzzyBrainwebPhantom, CrispPhantomMixin):
    """Crisp BrainWeb phantom builder."""

    def cache(self, file_path: str, array: np.ndarray):
        """
        Cache an array for fast retrieval.

        Parameters
        ----------
        file_path : str
            Path on disk to cached array.
        array : np.ndarray
            Array to be cached.

        """
        if os.path.exists(file_path) is False:
            array = _fuzzy_to_crisp(array)
            np.save(file_path, array)
            self.as_crisp(copy=False)


class NumericBrainwebPhantom(CrispBrainwebPhantom):
    """Numeric BrainWeb phantom builder."""

    def __init__(
        self,
        ndim: int,
        subject: int,
        shape: int | Sequence[int] = None,
        output_res: float | Sequence[float] = None,
        B0: float = 1.5,
        cache: bool = True,
        cache_dir: CacheDirType = None,
        brainweb_dir: CacheDirType = None,
        force: bool = False,
        verify: bool = True,
    ):

        super().__init__(
            ndim,
            subject,
            shape,
            output_res,
            B0,
            cache,
            cache_dir,
            brainweb_dir,
            force,
            verify,
        )

        self.as_numeric(copy=False)



"""Sub-package containing utility routines.

Resize
----------
These routines include array crop and pad.

Resampling
----------
These routines include array up- and downsampling.

Download
--------
Utilities for files download.

Typing
------
Custom data types for type hint.

Path
----
Utilities to handle i.e., cache folder position.


"""

__all__ = []

from . import _download
from . import _fft
from . import _pathlib
from . import _resample
from . import _resize
from . import _typing

from ._download import *  # noqa
from ._fft import *  # noqa
from ._pathlib import *  # noqa
from ._resample import *  # noqa
from ._resize import *  # noqa
from ._typing import *  # noqa

__all__.extend(_download.__all__)
__all__.extend(_fft.__all__)
__all__.extend(_pathlib.__all__)
__all__.extend(_resample.__all__)
__all__.extend(_resize.__all__)
__all__.extend(_typing.__all__)



"""Fields generation sub-package."""

__all__ = []

from ._b0_map import b0field  # noqa
from ._b1_map import b1field  # noqa
from ._sens_map import sensmap  # noqa

__all__.append("b0field")
__all__.append("b1field")
__all__.append("sensmap")



"""Phantom mixins."""

__all__ = ["PhantomMixin", "CrispPhantomMixin", "FuzzyPhantomMixin"]

import os

from copy import deepcopy

import numpy as np


class PhantomMixin:
    """Base phantom mixin."""

    def cache(self, file_path: str, array: np.ndarray):
        """
        Cache an array for fast retrieval.

        Parameters
        ----------
        file_path : str
            Path on disk to cached array.
        array : np.ndarray
            Array to be cached.

        """
        if os.path.exists(file_path) is False:
            np.save(file_path, array)


class CrispPhantomMixin(PhantomMixin):
    """Crisp phantom mixin."""

    def __getitem__(self, idx):  # noqa
        if self.segmentation is not None:
            return self.segmentation.__getitem__(idx)
        else:
            return None

    @property
    def shape(self):  # noqa
        if self.segmentation is not None:
            return self.segmentation.shape
        else:
            return None

    def as_numeric(self, copy: bool = True):  # noqa
        """Convert crisp phantom into numeric phantom."""
        if copy:
            out = deepcopy(self)
        else:
            out = self
        for param in out._properties.keys():
            param_map = np.zeros(out.segmentation.shape, dtype=np.float32)
            for idx in range(len(out._properties[param])):
                param_map += out._properties[param][idx] * (
                    out.segmentation == out._label[idx]
                )
            out._properties[param] = param_map

        return out


class FuzzyPhantomMixin(PhantomMixin):
    """Fuzzy phantom mixin."""

    def __getitem__(self, idx):  # noqa
        if self.segmentation is not None:
            return self.segmentation.__getitem__(idx)
        else:
            return None

    @property
    def shape(self):  # noqa
        if self.segmentation is not None:
            return self.segmentation.shape
        else:
            return None

    def as_crisp(self, copy: bool = True):
        """Convert fuzzy phantom into crisp phantom."""
        if copy:
            out = deepcopy(self)
        else:
            out = self
        out.segmentation = _fuzzy_to_crisp(out.segmentation)
        return out

    def as_numeric(self, copy: bool = True):
        """Convert fuzzy phantom into numeric phantom."""
        if self.segmentation.ndim != self._ndim:
            out = self.as_crisp(copy)
        else:
            if copy:
                out = deepcopy(self)
            else:
                out = self

        # build tissue maps
        for param in out._properties.keys():
            param_map = np.zeros(out.segmentation.shape, dtype=np.float32)
            for idx in range(len(out._properties[param])):
                param_map += out._properties[param][idx] * (
                    out.segmentation == out._label[idx]
                )
            out._properties[param] = param_map

        # erase segmentation
        out.segmentation = None

        return out


def _fuzzy_to_crisp(fuzzy_segmentation: np.ndarray) -> np.ndarray:
    """
    Convert fuzzy segmentation into crisp segmentation.

    Conversion is performed assigning each voxel to the class with
    highest probability.

    Parameters
    ----------
    fuzzy_segmentation : np.ndarray
        Input fuzzy segmentation of shape (nclasses, *shape).

    Returns
    -------
    np.ndarray
        Output crisp segmentation of shape (*shape).

    """
    crisp_segmentation = np.argmax(fuzzy_segmentation, axis=0)
    return crisp_segmentation.astype(int)



"""OSF Phantom sub-package."""

__all__ = ["osf_phantom"]

from typing import Sequence
from .._utils import CacheDirType, PhantomType

from ._osf import NumericOSFPhantom


def osf_phantom(
    ndim: int,
    subject: int,
    shape: int | Sequence[int] = None,
    output_res: float | Sequence[float] = None,
    B0: float = 3.0,
    cache: bool = True,
    cache_dir: CacheDirType = None,
    osf_dir: CacheDirType = None,
    force: bool = False,
    verify: bool = True,
) -> PhantomType:
    """
    Get OSF phantom.

    Parameters
    ----------
    ndim : int
        Number of spatial dimensions. If ndim == 2, use a single slice
        (central axial slice).
    subject : int
        Subject id to download.
    shape: int | Sequence[int] | None, optional
        Shape of the output data, the data will be interpolated to the given shape.
        If int, assume isotropic matrix. The default is ``None`` (original shape).
    output_res: float | Sequence[float] | None, optional
        Resolution of the output data, the data will be rescale to the given resolution.
        If scalar, assume isotropic resolution. The default is ``None``
        (estimate from shape assuming same fov).
    B0 : float, optional
        Static field strength in [T].
        The default is `3.0`.
    cache : bool, optional
        If ``True``, cache the phantom. The default is ``True``.
    cache_dir : CacheDirType, optional
        cache_directory for phantom caching.
        The default is ``None`` (``~/.cache/mrtwin``).
    osf_dir : CacheDirType, optional
        osf_directory for brainweb segmentation caching.
        The default is ``None`` (``~/.cache/osf``).
    force : bool, optional
        Force download even if the file already exists.
        The default is ``False``.
    verify : bool, optional
        Enable SSL verification.
        DO NOT DISABLE (i.e., ``verify=False``) IN PRODUCTION.
        The default is ``True``.

    Returns
    -------
    PhantomType
        OSF phantom.

    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from mrtwin import osf_phantom

    We can generate a dense single slice 2D phantom with a matrix size
    of ``(256, 256)`` at 1.0625 mm isotropic resolution for the
    OSF subject ``n=1`` as:

    >>> phantom = osf_phantom(ndim=2, subject=1)

    Phantom T1 and T2 maps, can be accessed as:

    >>> fig, ax = plt.subplots(2, 1)

    >>> im1 = ax[0].imshow(phantom.T1, cmap="magma", vmin=0, vmax=3500)
    >>> ax[0].axis("off"), ax[0].set_title("T1 [ms]")
    >>> fig.colorbar(im1, ax=ax[0], fraction=0.046, pad=0.04)

    >>> im2 = ax[1].imshow(phantom.T2, cmap="viridis", vmin=0, vmax=250)
    >>> ax[1].axis("off"), ax[1].set_title("T2 [ms]")
    >>> fig.colorbar(im, ax=ax[1], fraction=0.046, pad=0.04)

    """
    return NumericOSFPhantom(
        ndim,
        subject,
        shape,
        output_res,
        B0,
        cache,
        cache_dir,
        osf_dir,
        force,
        verify,
    )



"""
================================================
Open Science CBS Neuroimaging Repository Phantom
================================================

Example of Open Science CBS Neuroimaging Repository phantom creation.

This dataset consists of an additional processing of the dataset created
by Tardif et al. [1], and stored `in a Open Science Framework (OSF) repository <https://osf.io/qkbca/>`_.

This examples show how to generate numerical phantoms based on the Brainweb
dataset.

References
----------
[1] Tardif CL, Schäfer A, Trampel R, Villringer A, Turner R, Bazin PL. 
Open Science CBS Neuroimaging Repository: Sharing ultra-high-field MR images of the brain. 
Neuroimage. 2016 Jan 1;124(Pt B):1143-1148. doi: 10.1016/j.neuroimage.2015.08.042. 
Epub 2015 Aug 25. PMID: 26318051.

"""

import matplotlib.pyplot as plt
import numpy as np

from mrtwin import osf_phantom

# %%
# Basic Usage
# ===========
# The Open Science CBS Neuroimaging database (hereafter referred to as `OSF`)
# consists of a set of 20 normal brains (ids equal to ``[1-4, 6, 8-15, 17, 19, 22-23, 25, 27-28]``)
# at a nominal 0.5 mm isotropic and a shape of ``(nz, ny, nx) = (454, 544, 454)``.
#
# A digital OSF phantom can be created as:

phantom = osf_phantom(ndim=2, subject=1)  # 2D phantom

# %%
# Here, without loss of generality, we use a single-slice 2D phantom.
# A 3D phantom can be generated by setting ``ndim=3``.
#
# The ``(M0, T1, T2, T2*, Chi)`` properties of the phantom
# can be direcly accessed as:

fig1, ax1 = plt.subplots(1, 5)

im0 = ax1[0].imshow(phantom.M0, cmap="gray", vmin=0)
ax1[0].axis("off"), ax1[0].set_title("M0 [a.u.]")
fig1.colorbar(im0, ax=ax1[0], fraction=0.046, pad=0.04)

im1 = ax1[1].imshow(phantom.T1, cmap="magma", vmin=0)
ax1[1].axis("off"), ax1[1].set_title("T1 [ms]")
fig1.colorbar(im1, ax=ax1[1], fraction=0.046, pad=0.04)

im2 = ax1[2].imshow(phantom.T2, cmap="viridis", vmin=0, vmax=250)
ax1[2].axis("off"), ax1[2].set_title("T2 [ms]")
fig1.colorbar(im2, ax=ax1[2], fraction=0.046, pad=0.04)

im3 = ax1[3].imshow(phantom.T2s, cmap="viridis", vmin=0, vmax=250)
ax1[3].axis("off"), ax1[3].set_title("T2* [ms]")
fig1.colorbar(im3, ax=ax1[3], fraction=0.046, pad=0.04)

im4 = ax1[4].imshow(phantom.Chi, cmap="gray")
ax1[4].axis("off"), ax1[4].set_title("Chi")
fig1.colorbar(im4, ax=ax1[4], fraction=0.046, pad=0.04)

plt.tight_layout()
plt.show()

# %%
# If required, the ``properties`` dictionary can be directly accessed as ``phantom.properties``,
# e.g., to be passed as ``**kwargs`` to a simulator routine.

# %% Setting spatial properties
#
# By default, OSF phantoms are interpolated
# to 1.0625 mm isotropic resolution with a 256 isotropic matrix (272 mm iso FOV).
#
# These can be adjusted using `shape` and `output_res` model:
#
# 1. ``shape`` will control the matrix size without affecting the resolution.
# 2. ``output_res`` will adjust the spatial resolution keeping the same (i.e., 200 iso) matrix
#

phantom = osf_phantom(ndim=2, subject=1)
phantom_mtx = osf_phantom(ndim=2, subject=1, shape=200)  # can also be shape=(ny, nx)
phantom_res = osf_phantom(
    ndim=2, subject=1, output_res=2.0
)  # can also be output_res=(dy, dx)
phantom_mtx_res = osf_phantom(ndim=2, subject=1, shape=200, output_res=2.0)

print(phantom)
print(phantom_mtx)
print(phantom_res)
print(phantom_mtx_res)

fig2, ax2 = plt.subplots(2, 2)
ax2[0, 0].imshow(phantom.T1, cmap="magma"), ax2[0, 0].axis("off"), ax2[0, 0].set_title(
    "shape 256, res=1.0625, fov=272mm"
)
ax2[0, 1].imshow(phantom_mtx.T1, cmap="magma"), ax2[0, 1].axis("off"), ax2[
    0, 1
].set_title("shape 200, res=1.36mm, fov=272mm")
ax2[1, 0].imshow(phantom_res.T1, cmap="magma"), ax2[1, 0].axis("off"), ax2[
    1, 0
].set_title("shape 200, res=2.0mm, fov=400mm")
ax2[1, 1].imshow(phantom_mtx_res.T1, cmap="magma"), ax2[1, 1].axis("off"), ax2[
    1, 1
].set_title("shape 256, res=2.0mm, fov=512mm")
plt.show()


# %% Setting field strength
#
# The physical parameter of each tissue class are reported by
# default for a field strength of 3.0 T.
#
# This can be changed via the ``B0`` argument:

# B0 strengths
B0 = [0.55, 1.5, 3.0, 7.0, 11.7, 13.3]  # field strengths in [T]

# Generate phantoms with different field strengths
phantomB0 = [osf_phantom(ndim=2, subject=1, B0=strength) for strength in B0]

# Display
T1 = np.concatenate([phantom.T1 for phantom in phantomB0], axis=1)
T2 = np.concatenate([phantom.T2 for phantom in phantomB0], axis=1)
T2s = np.concatenate([phantom.T2s for phantom in phantomB0], axis=1)

fig5, ax5 = plt.subplots(3, 1)

im1 = ax5[0].imshow(T1, cmap="magma", vmin=0, vmax=5000)
ax5[0].axis("off"), ax5[0].set_title("T1 [ms]")
fig5.colorbar(im1, ax=ax5[0], fraction=0.046, pad=0.04)

im2 = ax5[1].imshow(T2, cmap="viridis", vmin=0, vmax=250)
ax5[1].axis("off"), ax5[1].set_title("T2 [ms]")
fig5.colorbar(im2, ax=ax5[1], fraction=0.046, pad=0.04)

im3 = ax5[2].imshow(T2s, cmap="viridis", vmin=0, vmax=250)
ax5[2].axis("off"), ax5[2].set_title("T2* [ms]")
fig5.colorbar(im3, ax=ax5[2], fraction=0.046, pad=0.04)

plt.tight_layout()
plt.show()

# %%
#
# In this case, T1 and T2* are extrapolated from their 3.0 T values.
#
# Caching mechanism
# =================
#
# To reduce loading times, ``mrtwin`` implements a caching mechanism.
#
# If ``cache`` argument is set to ``True`` (default behaviour), each phantom
# segmentation (identified by the number of spatial dimensions, matrix shape and resolution)
# is saved on the disk in ``npy`` format.
#
# The path is selected according to the following hierachy (inspired by ``brainweb-dl``):
#
# 1. User-specific argument (``cache_dir``)
# 2. ``MRTWIN_DIR`` environment variable
# 3. ``~/.cache/mrtwin`` folder
#
#
# Disabling SSL verification (not recommended)
# ============================================
#
# If you encounter some issue in downloading, SSL verification
# can be disabled by setting ``verify`` to ``False`` (default: ``True``).
# It is advised however to solve the problem on your machine side
# (updating the certificate).



"""Test Brainweb phantom generation."""

import os
import tempfile


import pytest


import numpy.testing as npt


from mrtwin import brainweb_phantom


@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("subject", [4, 44])
@pytest.mark.parametrize("shape", [None, 128, (128, 128), (128, 128, 128)])
@pytest.mark.parametrize("model", ["single-pool", "mw-model", "mt-model", "mwmt-model"])
@pytest.mark.parametrize("segtype", ["crisp", "fuzzy", False])
@pytest.mark.parametrize("output_res", [None, 1.0, (1.0, 1.0), (1.0, 1.0, 1.0)])
def test_brainweb_phantom(ndim, subject, shape, model, segtype, output_res):
    """
    Test the brainweb_phantom function with various parameter combinations.
    """
    if ndim == 2 and shape == (128, 128, 128):
        pytest.skip("2D phantom not compatible with len(shape)==3")
    if ndim == 3 and shape == (128, 128):
        pytest.skip("3D phantom not compatible with len(shape)==2")
    if ndim == 2 and output_res == (1.0, 1.0, 1.0):
        pytest.skip("2D phantom not compatible with len(output_res)==3")
    if ndim == 3 and output_res == (1.0, 1.0):
        pytest.skip("3D phantom not compatible with len(output_res)==2")

    phantom = brainweb_phantom(
        ndim=ndim,
        subject=subject,
        shape=shape,
        model=model,
        segtype=segtype,
        output_res=output_res,
        verify=False,
    )

    # Validate the output is not None
    assert phantom is not None, "Phantom output should not be None."

    # Validate the shape of the output if a shape is provided
    if shape is not None:
        expected_shape = tuple([shape] * ndim) if isinstance(shape, int) else shape
        actual_shape = phantom.shape[-ndim:] if segtype else phantom.T1.shape[-ndim:]
        npt.assert_allclose(actual_shape, expected_shape)


def test_brainweb_phantom_default():
    """
    Test the brainweb_phantom function with default parameters.
    """
    ndim = 2
    subject = 4

    phantom = brainweb_phantom(ndim=ndim, subject=subject, verify=False)

    # Validate the output is not None
    assert phantom is not None, "Phantom output should not be None."

    # Validate the default shape for 2D phantom (since no shape was provided)
    assert phantom.shape[-2:] == (200, 200)



"""Birdcage map generation."""

__all__ = ["_birdcage"]

import math


import numpy as np


def _birdcage(shape, coil_width, nrings, shift, dphi):  # noqa
    # default
    if shift is None:
        shift = [0.0 for ax in range(len(shape) - 1)]
    if nrings is None:
        nrings = np.max((shape[0] // 4, 1))

    # coil width and radius
    c_width = coil_width * min(shape[-2:])
    c_rad = 0.5 * c_width

    if len(shape) == 3:
        nc, ny, nx = shape
        phi = np.arange(nc) * (2 * math.pi / nc) + dphi
        y, x = np.mgrid[:ny, :nx]

        x0 = c_rad * np.cos(phi) + shape[-1] / 2.0 + shift[-1]
        y0 = c_rad * np.sin(phi) + shape[-2] / 2.0 + shift[-2]

        x_co = x[None, ...] - x0[:, None, None]
        y_co = y[None, ...] - y0[:, None, None]

        # coil magnitude
        rr = np.sqrt(x_co**2 + y_co**2) / (2 * c_width)

        # coil phase
        phi = np.arctan2(x_co, -y_co) - phi[:, None, None]

    elif len(shape) == 4:
        nc, nz, ny, nx = shape
        phi = np.arange(nc) * (2 * math.pi / (nc + nrings)) + dphi
        z, y, x = np.mgrid[:nz, :ny, :nx]

        x0 = c_rad * np.cos(phi) + shape[-1] / 2.0 + shift[-1]
        y0 = c_rad * np.sin(phi) + shape[-2] / 2.0 + shift[-2]
        z0 = (
            np.floor(np.arange(nc) / nrings)
            - 0.5 * (np.ceil(np.arange(nc) / nrings) - 1)
            + shape[-3] / 2.0
            + shift[-3]
        )

        x_co = x[None, ...] - x0[:, None, None, None]
        y_co = y[None, ...] - y0[:, None, None, None]
        z_co = z[None, ...] - z0[:, None, None, None]

        # coil magnitude
        rr = np.sqrt(x_co**2 + y_co**2 + z_co**2) / (2 * c_width)

        # coil phase
        phi = np.arctan2(x_co, -y_co) - phi[:, None, None, None]
    else:
        raise ValueError("Can only generate shape with length 3 or 4")

    # build coils
    rr[rr == 0.0] = 1.0
    smap = (1.0 / rr) * np.exp(1j * phi)

    return smap.astype(np.complex64)



"""B1+ field maps generation routines."""

__all__ = ["b1field"]


import math
import os


from typing import Sequence


import numpy as np


from .._utils import CacheDirType, get_mrtwin_dir


from ._birdcage import _birdcage


def b1field(
    shape: Sequence[int],
    nmodes: int = 1,
    b1range: Sequence[float] = (0.5, 2.0),
    shift: Sequence[int] | None = None,
    dphi: float = 0.0,
    coil_width: float = 1.1,
    ncoils: int = 2,
    nrings: int | None = None,
    mask: np.ndarray | None = None,
    cache: bool | None = None,
    cache_dir: CacheDirType = None,
):
    """
    Simulate inhomogeneous B1+ fields.

    Adapted from SigPy [1].

    Parameters
    ----------
    shape : Sequence[int]
        Size of the matrix ``(ny, nx)`` (2D) or
        ``(nz, ny, nx)`` (3D) for the B1+ field.
    nmodes : int, optional
        Number of B1+ modes. First mode is ``CP`` mode, second
        is ``gradient`` mode, and so on. The default is ``1``.
    b1range : Sequence[float], optional
        Range of B1+ magnitude. The default is ``(0.5, 2.0)``.
    shift : Sequence[int], optional
        Displacement of the coil center with respect to matrix center.
        The default is ``(0, 0)`` / ``(0, 0, 0)``.
    dphi : float, optional
        Bulk coil angle in ``[deg]``.
        The default is ``0.0°``.
    coil_width : float, optional
        Width of the coil, with respect to image dimension.
        The default is ``1.1``.
    ncoils : int, optional
        Number of transmit coil elements. Standard coils have ``2`` channels
        operating in quadrature. To support multiple modes (e.g., PTX), increase this
        number. The default is ``4``.
    nrings : int, optional
        Number of rings for a cylindrical hardware set-up.
        The default is ``ncoils // 4``.
    mask : np.ndarray | None, optional
        Region of support of the object of
        shape ``(ny, nx)`` (2D) or ``(nz, ny, nx)`` (3D).
        The default is ``None``.
    cache : bool | None, optional
        If ``True``, cache the phantom.
        The default is ``True`` for 3D phantoms
        and ``False`` for single-slice 2D.
    cache_dir : CacheDirType, optional
        cache_directory for phantom caching.
        The default is ``None`` (``~/.cache/mrtwin``).

    Returns
    -------
    smap : np.ndarray
        Complex spatially varying b1+ maps of shape ``(nmodes, ny, nx)`` (2D)
        or ``(nmodes, nz, ny, nx)`` (3D). Magnitude of the map represents
        the relative flip angle scaling (wrt to the nominal).

    Example
    -------
    >>> from mrtwin import b1field

    We can generate a 2D B1+ field map of shape ``(ny=128, nx=128)`` by:

    >>> b1map = b1field((128, 128))

    Field center and rotation can be modified by ``shift`` and ``dphi`` arguments:

    >>> b1map = b1field((128, 128), shift=(-3, 5), dphi=30.0) # center shifted by (dy, dx) = (-3, 5) pixels and rotated by 30.0 degrees.

    B1+ values range and steepness of variation can be specified using ``b1range`` and ``coil_width`` arguments:

    >>> # transmit coil is 4 times bigger than FOV (e.g., body coil) and
    >>> # B1+ scalings are between (0.8, 1.2) the nominal flip angle (e.g., 3T scanner)
    >>> b1map3T = b1field((128, 128), b1range=(0.8, 1.2), coil_width=4.0)
    >>>
    >>> # transmit coil is 1.1 times bigger than FOV (e.g., head coil) and
    >>> # B1+ scalings are between (0.5, 2.0) the nominal flip angle (e.g., 7T scanner)
    >>> b1map7T = b1field((128, 128), b1range=(0.5, 2.0), coil_width=1.1)

    Multiple orthogonal modes can be simulated by ``nmodes`` argument.
    For example, `CP` mode and `gradient` mode can be obtained as:

    >>> b1map = b1field((128, 128), nmodes=2) # b1map[0] is CP, b1map[1] is gradient mode.

    Three dimensional B1+ maps of shape ``(nz, ny, nx)`` can be obtained as:

    >>> b1map = b1field((128, 128, 128))

    Beware that this will require more memory.

    References
    ----------
    [1] https://github.com/mikgroup/sigpy/tree/main

    """
    ncoils *= 2

    # check we can do quadrature
    assert (
        ncoils >= 2
    ), f"We support circular polarization only - found {ncoils} transmit elements."
    assert ncoils >= nmodes, f"Need ncoils (={ncoils}) to be >= nmodes (={nmodes})."

    # Default values
    if shift is None:
        shift = [0.0 for ax in range(len(shape))]
    if nrings is None:
        nrings = np.max((shape[0] // 4, 1))
    if cache is None and len(shape) == 2:  # (ny, nx) -> 2D
        cache = False
    elif cache is None and len(shape) == 3:  # (nz, ny, nx) -> 3D
        cache = True

    # Get filename for caching
    shape_str = [str(di) for di in shape]
    shape_str = "x".join(tuple(shape_str))

    shift_str = [str(di) for di in shift]
    shift_str = "x".join(tuple(shift_str))

    b1range_str = [str(value) for value in b1range]
    b1range_str = "-".join(tuple(b1range_str))

    if mask is None:
        is_masked = False
    else:
        is_masked = True

    file_name = f"b1map{shape_str}mtx_{nmodes}modes_{b1range_str}range_{coil_width}width_{ncoils}coils_{shift_str}shift_{nrings}rings_{dphi}deg_{is_masked}.npy"

    # Get base directory
    cache_dir = get_mrtwin_dir(cache_dir)

    # Get file path
    file_path = os.path.join(cache_dir, file_name)

    # Try to load
    if os.path.exists(file_path):
        return np.load(file_path)

    # Generate coils
    smap = _birdcage(
        [ncoils] + list(shape), coil_width, nrings, shift, np.deg2rad(dphi)
    )

    # Normalize
    rss = sum(abs(smap) ** 2, 0) ** 0.5
    smap /= rss

    # Combine
    dalpha = 2 * math.pi / ncoils
    alpha = np.arange(ncoils) * dalpha
    mode = np.arange(nmodes)
    phafu = np.exp(1j * mode[:, None] * alpha[None, :])  # (nmodes, nchannels)

    # Get modes
    smap = smap.T  # (nc, ...) -> (..., nc)
    smap = [(abs(smap) * phafu[n]).sum(axis=-1) for n in range(nmodes)]
    smap = np.stack(smap, axis=-1)  # (..., nmodes)
    smap = smap.T  # (..., nmodes) -> (nmodes, ...)

    # Rescale
    phase = smap / abs(smap)
    smap = abs(smap)
    smap = smap - smap.min()  # (min, max) -> (0, max - min)
    smap = smap / smap.max()  # (0, max - min) -> (0, 1)
    smap = (
        smap * (b1range[1] - b1range[0]) + b1range[0]
    )  # (0, 1) -> (b1range[0], b1range[1])
    smap = smap * phase

    if nmodes == 1:
        smap = abs(smap[0]).astype(np.float32)
    else:
        smap = smap.astype(np.complex64)

    # Mask
    if mask is not None:
        smap = mask * smap

    # Cache the result
    if cache and os.path.exists(file_path) is False:
        np.save(file_path, smap)

    return smap



"""Array resampling routines."""

__all__ = ["resample"]

import numpy as np


from scipy.ndimage import zoom


from ._broadcasting import _expand_shapes


def resample(input, oshape):
    """
    Resample a n-dimensional signal.

    Parameters
    ----------
    input : np.ndarray
        Input array of shape ``(..., ishape)``.
    oshape : Sequence
        Output shape.

    Returns
    -------
    output : np.ndarray
        Resampled tensor of shape ``(..., oshape)``.

    """
    if isinstance(oshape, int):
        oshape = [oshape]

    # get initial and final shapes
    ishape1, oshape1 = _expand_shapes(input.shape, oshape)

    # interpolate
    scale = np.asarray(oshape1) / np.asarray(ishape1)

    return zoom(input, scale, order=1)



"""Test static field map generation."""

import pytest
import numpy as np

from mrtwin import b0field


@pytest.mark.parametrize("shape", [(128, 128), (64, 64), (64, 64, 64)])
def test_b0field_basic_shape(shape):
    """
    Test the b0field function with basic input shapes for chi (susceptibility map).
    """
    # Create a random susceptibility map (chi) of the given shape
    chi = np.random.rand(*shape)

    # Call the b0field function
    b0map = b0field(chi)

    # Validate the output is a numpy ndarray
    assert isinstance(b0map, np.ndarray), "B0 map output should be a numpy ndarray."

    # Validate the output shape matches the input shape
    assert b0map.shape == shape, f"Expected shape {shape}, but got {b0map.shape}."


@pytest.mark.parametrize("b0range", [None, (-500, 500), (-1000, 1000)])
def test_b0field_b0range(b0range):
    """
    Test the b0field function with different b0range values.
    """
    shape = (128, 128)
    chi = np.random.rand(*shape)

    # Call the b0field function with varying b0range
    b0map = b0field(chi, b0range=b0range)

    # Validate the output is a numpy ndarray
    assert isinstance(b0map, np.ndarray), "B0 map output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert b0map.shape == shape, f"Expected shape {shape}, but got {b0map.shape}."

    if b0range:
        # Check that values are within the specified range
        assert np.all(b0map >= b0range[0]) and np.all(b0map <= b0range[1]), (
            f"B0 map values should be in range {b0range}, but got min {b0map.min()} "
            f"and max {b0map.max()}"
        )


def test_b0field_default_b0range():
    """
    Test the b0field function with default b0range and a static field B0 of 1.5T.
    """
    shape = (128, 128)
    chi = np.random.rand(*shape)

    # Call the b0field function without specifying b0range (default)
    b0map = b0field(chi, B0=1.5)

    # Validate the output is a numpy ndarray
    assert isinstance(b0map, np.ndarray), "B0 map output should be a numpy ndarray."

    # Validate the output shape matches the input shape
    assert b0map.shape == shape, f"Expected shape {shape}, but got {b0map.shape}."


@pytest.mark.parametrize("mask", [None, np.ones((128, 128)), np.zeros((128, 128))])
def test_b0field_with_mask(mask):
    """
    Test the b0field function with and without a mask.
    """
    shape = (128, 128)
    chi = np.random.rand(*shape)

    # Call the b0field function with the mask
    b0map = b0field(chi, mask=mask)

    # Validate the output is a numpy ndarray
    assert isinstance(b0map, np.ndarray), "B0 map output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert b0map.shape == shape, f"Expected shape {shape}, but got {b0map.shape}."


def test_b0field_with_3d_input():
    """
    Test the b0field function with 3D input data.
    """
    shape = (64, 64, 64)
    chi = np.random.rand(*shape)

    # Call the b0field function with 3D input
    b0map = b0field(chi)

    # Validate the output is a numpy ndarray
    assert isinstance(b0map, np.ndarray), "B0 map output should be a numpy ndarray."

    # Validate the output shape matches the input shape
    assert b0map.shape == shape, f"Expected shape {shape}, but got {b0map.shape}."


def test_b0field_output_units():
    """
    Test that the b0field output has units in Hz by checking magnitude with different B0 values.
    """
    shape = (128, 128)
    chi = np.random.rand(*shape)

    # Compute B0 maps with different static field strengths
    b0map_1T = b0field(chi, B0=1.0)
    b0map_3T = b0field(chi, B0=3.0)

    # Ensure the output is still a numpy ndarray
    assert isinstance(b0map_1T, np.ndarray), "B0 map output should be a numpy ndarray."
    assert isinstance(b0map_3T, np.ndarray), "B0 map output should be a numpy ndarray."

    # Ensure the shapes are correct
    assert b0map_1T.shape == shape, f"Expected shape {shape}, but got {b0map_1T.shape}."
    assert b0map_3T.shape == shape, f"Expected shape {shape}, but got {b0map_3T.shape}."

    # Check that B0 scaling is applied (3T should have larger field values than 1T)
    assert np.max(np.abs(b0map_3T)) > np.max(
        np.abs(b0map_1T)
    ), "B0 field with 3T should have larger field values than with 1T."



"""
===================
Shepp-Logan Phantom
===================

Example of Shepp-Logan phantom creation.

This examples show how to generate numerical phantoms based on the 
Shepp-Logan model.

"""

import matplotlib.pyplot as plt
import numpy as np

from mrtwin import shepplogan_phantom

# %%
# Basic Usage
# ===========
# A digital Shepp-Logan phantom can be created as:

phantom2D = shepplogan_phantom(ndim=2, shape=200)  # 2D phantom
phantom3D = shepplogan_phantom(ndim=3, shape=200)  # 3D phantom

# %%
# where ``shape`` is a matrix size, which can be provided as a scalar (isotropic matrix)
# or a ``ndim``-length tuple.
#
# The phantoms here created are sparse, i.e., they consists of an integer
# ``(*spatial_shape)`` shaped ``np.ndarray`` whose values represent the label
# (i.e., the index) identifying each tissue type (e.g., Gray Matter, White Matter, CSF)
# and a list of ``(nclasses,)`` dictionaries each containing the ``(M0, T1, T2, T2*, Chi)``
# values for each class:

example2D = np.concatenate((phantom2D, phantom2D, phantom2D), axis=0)

example3Dax = np.concatenate((phantom3D[100], phantom3D[100], phantom3D[100]), axis=0)
example3Dcor = np.concatenate(
    (phantom3D[::-1, 100], phantom3D[::-1, 100], phantom3D[::-1, 100]), axis=0
)
example3Dsag = np.concatenate(
    (
        phantom3D[::-1, :, 100],
        phantom3D[::-1, :, 100],
        phantom3D[::-1, :, 100],
    ),
    axis=0,
)
example3D = np.concatenate((example3Dax, example3Dcor, example3Dsag), axis=1)

fig1, ax1 = plt.subplots(1, 2)
ax1[0].imshow(example2D, cmap="turbo"), ax1[0].axis("off"), ax1[0].set_title(
    "2D phantom"
)
ax1[1].imshow(example3D, cmap="turbo"), ax1[1].axis("off"), ax1[1].set_title(
    "3D phantom"
)
plt.show()

# %%
# The ``(M0, T1, T2, T2*, Chi)`` properties
# can be direcly accessed as:

_, _ = print("M0:", end="\t"), print(phantom2D.M0)  # same for phantom3D
_, _ = print("T1 (ms):", end="\t"), print(phantom2D.T1)
_, _ = print("T2 (ms):", end="\t"), print(phantom2D.T2)
_, _ = print("T2* (ms):", end="\t"), print(phantom2D.T2s)
_, _ = print("Chi:", end="\t"), print(phantom2D.Chi)

# %%
# If required, the ``properties`` dictionary
# can be directly accessed as:

print(phantom2D.properties)

# %%
# e.g., to be passed as ``**kwargs`` to a simulator routine.
#
# Notice that segmentation can be accessed directly (in read-only mode)
# via square bracked indexing, similarly to numpy arrays.
#
# A basic summary of the properties can be accessed
# via the ``__repr__`` attribute (i.e., enabling pretty printing):

print(phantom2D)
print(phantom3D)

# %%
# We can obtain a "dense" phantom,
# i.e., an object without segmentation whose
# ``(M0, T1, T2, T2*, Chi)`` properties are stored
# as parametric maps rather than the individual values
# of each tissue class as:

phantom2D = phantom2D.as_numeric()

# Print summary

print(phantom2D)

# Display parameter maps

fig3, ax3 = plt.subplots(1, 5)

im0 = ax3[0].imshow(phantom2D.M0, cmap="gray")
ax3[0].axis("off"), ax3[0].set_title("M0 [a.u.]")
fig3.colorbar(im0, ax=ax3[0], fraction=0.046, pad=0.04)

im1 = ax3[1].imshow(phantom2D.T1, cmap="magma")
ax3[1].axis("off"), ax3[1].set_title("T1 [ms]")
fig3.colorbar(im1, ax=ax3[1], fraction=0.046, pad=0.04)

im2 = ax3[2].imshow(phantom2D.T2, cmap="viridis", vmax=150)
ax3[2].axis("off"), ax3[2].set_title("T2 [ms]")
fig3.colorbar(im2, ax=ax3[2], fraction=0.046, pad=0.04)

im3 = ax3[3].imshow(phantom2D.T2s, cmap="viridis", vmax=150)
ax3[3].axis("off"), ax3[3].set_title("T2* [ms]")
fig3.colorbar(im3, ax=ax3[3], fraction=0.046, pad=0.04)

im4 = ax3[4].imshow(phantom2D.Chi, cmap="gray")
ax3[4].axis("off"), ax3[4].set_title("Chi")
fig3.colorbar(im4, ax=ax3[4], fraction=0.046, pad=0.04)

plt.tight_layout()
plt.show()

# %%
# Dense phantom can be also directly generated as:

phantom2D = shepplogan_phantom(
    ndim=2, shape=200, segtype=False
)  # default is segtype="crisp"

# Print summary

print(phantom2D)

# %%
# Hereafter, without loss of generality, we will use 2D phantoms.

# %% Setting field strength
#
# The physical parameter of each tissue class are calculated by
# default for a field strength of 1.5 T.
#
# This can be changed via the ``B0`` argument:

# B0 strengths
B0 = [0.55, 1.5, 3.0, 7.0, 11.7, 13.3]  # field strengths in [T]

# Generate phantoms with different field strengths
phantomB0 = [
    shepplogan_phantom(ndim=2, shape=200, B0=strength, segtype=False) for strength in B0
]

# Display
T1 = np.concatenate([phantom.T1 for phantom in phantomB0], axis=1)
T2 = np.concatenate([phantom.T2 for phantom in phantomB0], axis=1)
T2s = np.concatenate([phantom.T2s for phantom in phantomB0], axis=1)

fig5, ax5 = plt.subplots(3, 1)

im1 = ax5[0].imshow(T1, cmap="magma", vmax=5000)
ax5[0].axis("off"), ax5[0].set_title("T1 [ms]")
fig5.colorbar(im1, ax=ax5[0], fraction=0.046, pad=0.04)

im2 = ax5[1].imshow(T2, cmap="viridis", vmax=150)
ax5[1].axis("off"), ax5[1].set_title("T2 [ms]")
fig5.colorbar(im2, ax=ax5[1], fraction=0.046, pad=0.04)

im3 = ax5[2].imshow(T2s, cmap="viridis", vmax=100)
ax5[2].axis("off"), ax5[2].set_title("T2* [ms]")
fig5.colorbar(im3, ax=ax5[2], fraction=0.046, pad=0.04)

plt.tight_layout()
plt.show()


# %% Advanced tissue models
#
# In addition to single pool model, we provide 3 multi-pool models:
#
# 1. ``"mw-model"``: a two-pool model where free water is divided in two compartments,
#    i.e., `intra-/extra-cellular water` (long T1 / T2) and `myelin water` (short T1 / T2).
#    The model include chemical exchange between the two pools.
#    Parameters are ``(MWF, T1, T2, k, chemshift)``.
# 2. ``"mt-model"``: a two-pool model consisting in `free water` and `bound water.`
#    Free water includes both intra-/extra-cellular and myelin water (as in the single-pool model),
#    while bound water corresponds to a macromolecular pool with the same T1 as the free water
#    and no T2 (i.e., no transverse magnetization).
#    The model include magnetization transfer between the two pools.
#    Parameters are ``(MVF, T1, T2, k)``.
# 3. ``"mwmt-model"``: a three-pool model consisting in `intra-/extra-cellular water`, `myelin water` and `bound water`.
#    The model include chemical exchange between the two free water pools and magnetizion transfer between
#    myelin water and bound water.
#    Parameters are ``(MWF, MVF, T1, T2, k)``.
#
# Here we will display the latter, as it represents the most general case.

# model="single-pool" is the default, while "mw-model" and "mt-model" corresponds to cases 1. and 2.
phantom_multi = shepplogan_phantom(ndim=2, shape=200, model="mwmt-model", segtype=False)

# %%
#
# MWF corresponds to the myelin water fraction, while MVF to the bound water fraction.
# We assume that intra-extracellular water fraction ``= 1 - (MWF + MVF)``:

MWF = phantom_multi.MWF
MVF = phantom_multi.MVF
IEWF = (1 - (MWF + MVF)) * (MWF > 0)
weight = np.concatenate((IEWF, MWF, MVF), axis=1)

plt.figure()
plt.imshow(weight, vmin=0, vmax=1, cmap="hot"), plt.axis("off"), plt.title(
    "pool fractions"
), plt.colorbar()
plt.show()

# %%
#
# T1 and T2 for the two free water pools are stacked along the first axis,
# with ``n=0`` being the intra-/extra-cellular water (long T1 / T2) and
# ``n=1`` being the myelin water (short T1 / T2):

T1 = np.concatenate((phantom_multi.T1[0], phantom_multi.T1[1]), axis=1)
T2 = np.concatenate((phantom_multi.T2[0], phantom_multi.T2[1]), axis=1)

fig6, ax6 = plt.subplots(2, 1)

im1 = ax6[0].imshow(T1, cmap="magma", vmax=1500)
ax6[0].axis("off"), ax6[0].set_title("T1 [ms]")
fig6.colorbar(im1, ax=ax6[0], fraction=0.046, pad=0.04)

im2 = ax6[1].imshow(T2, cmap="viridis", vmax=150)
ax6[1].axis("off"), ax6[1].set_title("T2 [ms]")
fig6.colorbar(im2, ax=ax6[1], fraction=0.046, pad=0.04)


# %%
#
# k represent the non-directional exchange rates in [Hz],
# with ``n=0`` being the chemical exchange rate between the two free water pools
# and ``n=1`` being magnetization transfer rate between the myelin and bound water:

k = np.concatenate((phantom_multi.k[0], phantom_multi.k[1]), axis=1)

plt.figure()
plt.imshow(k, cmap="hot"), plt.axis("off"), plt.title(
    "exchange rate [Hz]"
), plt.colorbar()
plt.show()

# %%
#
# Similarly to the single pool model, ``mrtwin`` also supports a sparse (``segtype="crisp"``) representation.
#
# Caching mechanism
# =================
#
# To reduce loading times, ``mrtwin`` implements a caching mechanism.
#
# If ``cache`` argument is set to ``True`` (default behaviour for ``ndim=3``), each phantom
# segmentation (identified by the number of spatial dimensions,
# tissue model, segmentation type and matrix shape)
# is saved on the disk in ``npy`` format.
#
# The path is selected according to the following hierachy (inspired by ``brainweb-dl``):
#
# 1. User-specific argument (``cache_dir``)
# 2. ``MRTWIN_DIR`` environment variable
# 3. ``~/.cache/mrtwin`` folder
#



from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from sphinx_gallery.notebook import add_code_cell, add_markdown_cell

import os
import json


class ColabLinkNode(nodes.General, nodes.Element):
    """A custom docutils node to represent the Colab link."""


def visit_colab_link_node_html(self, node):
    self.body.append(node["html"])


def depart_colab_link_node_html(self, node):
    pass


class ColabLinkDirective(SphinxDirective):
    """Directive to insert a link to open a notebook in Google Colab."""

    has_content = True
    option_spec = {
        "needs_gpu": int,
    }

    def run(self):
        """Run the directive."""
        # Determine the path of the current .rst file
        rst_file_path = self.env.doc2path(self.env.docname)
        rst_file_dir = os.path.dirname(rst_file_path)

        # Determine the notebook file path assuming it is in the same directory as the .rst file
        notebook_filename = os.path.basename(rst_file_path).replace(".rst", ".ipynb")

        # Full path to the notebook
        notebook_full_path = os.path.join(rst_file_dir, notebook_filename)

        # Convert the full path back to a relative path from the repo root
        # repo_root = self.config.project_root_dir
        notebook_repo_relative_path = os.path.relpath(
            notebook_full_path, os.path.join(os.getcwd(), "docs")
        )

        # Generate the Colab URL based on GitHub repo information
        self.colab_url = f"https://colab.research.google.com/github/mind-inria/mri-nufft/blob/gh-pages/examples/{notebook_repo_relative_path}"

        # Create the HTML button or link
        self.html = f"""<div class="colab-button">
            <a href="{self.colab_url}" target="_blank">
                <img src="https://colab.research.google.com/assets/colab-badge.svg" 
                alt="Open In Colab"/>
            </a>
        </div>
        """
        self.notebook_modifier(notebook_full_path, "\n".join(self.content))

        # Create the node to insert the HTML
        node = ColabLinkNode(html=self.html)
        return [node]

    def notebook_modifier(self, notebook_path, commands):
        """Modify the notebook to add a warning about GPU requirement."""
        with open(notebook_path) as f:
            notebook = json.load(f)
        if "cells" not in notebook:
            notebook["cells"] = []

        # Add a cell to install the required libraries at the position where we have
        # colab link
        idx = self.find_index_of_colab_link(notebook)
        code_lines = ["# Install libraries"]
        code_lines.append(commands)
        dummy_notebook_content = {"cells": []}
        add_code_cell(
            dummy_notebook_content,
            "\n".join(code_lines),
        )
        notebook["cells"][idx] = dummy_notebook_content["cells"][0]

        needs_GPU = self.options.get("needs_gpu", False)
        if needs_GPU:
            # Add a warning cell at the top of the notebook
            warning_template = "\n".join(
                [
                    "<div class='alert alert-{message_class}'>",
                    "",
                    "# Need GPU warning",
                    "",
                    "{message}",
                    "</div>",
                    self.html,
                ]
            )
            message_class = "warning"
            message = (
                "Running this mri-nufft example requires a GPU, and hence is NOT "
                "possible on binder currently We request you to kindly run this notebook "
                "on Google Colab by clicking the link below. Additionally, please make "
                "sure to set the runtime on Colab to use a GPU and install the below "
                "libraries before running."
            )
            idx = 0
        else:
            # Add a warning cell at the top of the notebook
            warning_template = "\n".join(
                [
                    "<div class='alert alert-{message_class}'>",
                    "",
                    "# Install libraries needed for Colab",
                    "",
                    "{message}",
                    "</div>",
                    self.html,
                ]
            )
            message_class = "info"
            message = (
                "The below installation commands are needed to be run only on "
                "Google Colab."
            )

        dummy_notebook_content = {"cells": []}
        add_markdown_cell(
            dummy_notebook_content,
            warning_template.format(message_class=message_class, message=message),
        )
        notebook["cells"] = (
            notebook["cells"][:idx]
            + dummy_notebook_content["cells"]
            + notebook["cells"][idx:]
        )

        # Write back updated notebook
        with open(notebook_path, "w", encoding="utf-8") as f:
            json.dump(notebook, f, ensure_ascii=False, indent=2)

    def find_index_of_colab_link(self, notebook):
        """Find the index of the cell containing the Colab link."""
        for idx, cell in enumerate(notebook["cells"]):
            if cell["cell_type"] == "markdown" and ".. colab-link::" in "".join(
                cell.get("source", "")
            ):
                return idx
        return 0


def setup(app):
    """Set up the Sphinx extension."""
    app.add_node(
        ColabLinkNode, html=(visit_colab_link_node_html, depart_colab_link_node_html)
    )
    app.add_directive("colab-link", ColabLinkDirective)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Base Shepp-Logan phantom builder class."""

__all__ = ["SheppLoganPhantom"]

import os
import numpy as np

from typing import Sequence

from .._build import PhantomMixin
from .._utils import CacheDirType, get_mrtwin_dir

from ._segmentation import get_shepp_logan


class SheppLoganPhantom(PhantomMixin):
    """Base Shepp-Logan phantom builder."""

    def __init__(
        self,
        ndim: int,
        shape: int | Sequence[int] | None = None,
        cache: bool = True,
        cache_dir: CacheDirType = None,
    ):
        # keep dim
        self._ndim = ndim

        # default shape
        if np.isscalar(shape):
            shape = [shape] * ndim
        shape = np.asarray(shape)

        # get filename
        _fname = self.get_filename(ndim, shape)

        # try to load segmentation
        self.segmentation, file_path = self.get_segmentation(
            _fname,
            ndim,
            shape,
            cache,
            cache_dir,
        )

        # cache the result
        if cache:
            self.cache(file_path, self.segmentation)

    def __repr__(self):  # noqa
        if self.segmentation is None:
            ptype = "Dense"
        elif len(self.segmentation.shape) == self._ndim:
            ptype = "Crisp"
        msg = f"{ptype} Shepp-Logan phantom with following properties:\n"
        msg += f"Number of spatial dimensions: {self._ndim}\n"
        msg += f"Tissue properties: {self._properties.keys()}\n"
        if self.segmentation is not None:
            _shape = self.shape[-self._ndim :]
        else:
            _shape = list(self._properties.values())[0].shape[-self._ndim :]
        msg += f"Matrix size: {_shape}\n"
        if self.segmentation is not None and len(self.segmentation.shape) != self._ndim:
            msg += f"Number of tissue classes: {self.segmentation.shape[0]}\n"
        return msg

    def get_filename(
        self,
        ndim: int,
        shape: int | Sequence[int],
    ):
        """
        Generate filename starting from matrix shape.

        Parameters
        ----------
        ndim : int
            Number of spatial dimensions. If ndim == 2, use a single slice
            (central axial slice).
        shape: int | Sequence[int]
            Shape of the output data, the data will be interpolated to the given shape.
            If int, assume isotropic matrix.

        Returns
        -------
        str
            Filename for caching.

        """
        shape_str = [str(di) for di in shape.tolist()]
        shape_str = "x".join(tuple(shape_str))

        return f"{self.__class__.__name__.lower()}_{shape_str}mtx.npy"

    def get_segmentation(
        self,
        fname: str,
        ndim: int,
        shape: int | Sequence[int],
        cache: bool,
        cache_dir: CacheDirType,
    ):
        """
        Get crisp Shepp-Logan tissue segmentation.

        Parameters
        ----------
        fname : str
            Filename for caching.
        ndim : int
            Number of spatial dimensions. If ndim == 2, use a single slice
            (central axial slice).
        shape: int | Sequence[int]
            Shape of the output data, the data will be interpolated to the given shape.
            If int, assume isotropic matrix.
        cache : bool
            If True, cache the result.
        cache_dir : CacheDirType
            Directory for segmentation caching.

        Returns
        -------
        np.ndarray.
            Shepp-Logan segmentation.
        file_path : str
            Path on disk to generated segmentation for caching.

        """
        # get base directory
        cache_dir = get_mrtwin_dir(cache_dir)

        # get file path
        file_path = os.path.join(cache_dir, fname)

        # try to load
        if os.path.exists(file_path):
            return np.load(file_path), file_path
        else:
            segmentation = get_shepp_logan(ndim, shape)

        return segmentation, file_path

    def __array__(self):  # noqa
        # This method tells NumPy how to convert the object to an array
        return self.segmentation



"""Two-pool (I/E W + MW) Shepp-Logan phantom builder class."""

__all__ = ["CrispMWSheppLoganPhantom", "NumericMWSheppLoganPhantom"]

import numpy as np

from .. import _classes

from ._shepplogan import CrispSheppLoganPhantom, NumericSheppLoganPhantom


class CrispMWSheppLoganPhantom(CrispSheppLoganPhantom):
    """Crisp MW Shepp-Logan phantom builder."""

    def get_model(self, B0: float):
        """Initialize model.

        Parameters
        ----------
        B0, float
            Static field strength in [T].

        """
        model_dict = _classes.tissue_map("mw-model")
        self._label = []
        self._properties = {
            "MWF": [],
            "T1w": [],
            "T1m": [],
            "T2w": [],
            "T2m": [],
            "k": [],
            "chemshift": [],
        }
        for tissue in model_dict:
            self._label.append(tissue["Label"])
            self._properties["MWF"].append(tissue["MWF"])
            self._properties["T1w"].append(tissue["T1w"])
            self._properties["T1m"].append(tissue["T1m"])
            self._properties["T2w"].append(tissue["T2w"])
            self._properties["T2m"].append(tissue["T2m"])
            self._properties["k"].append(tissue["k"])
            self._properties["chemshift"].append(tissue["chemshift"])

        # cast to array
        self._label = np.asarray(self._label, dtype=int)
        self._properties["MWF"] = np.asarray(self._properties["MWF"], dtype=np.float32)
        self._properties["T1w"] = np.asarray(self._properties["T1w"], dtype=np.float32)
        self._properties["T1m"] = np.asarray(self._properties["T1m"], dtype=np.float32)
        self._properties["T2w"] = np.asarray(self._properties["T2w"], dtype=np.float32)
        self._properties["T2m"] = np.asarray(self._properties["T2m"], dtype=np.float32)
        self._properties["k"] = np.asarray(self._properties["k"], dtype=np.float32)
        self._properties["chemshift"] = np.asarray(
            self._properties["chemshift"], dtype=np.float32
        )

    @property
    def MWF(self):  # noqa
        return self._properties["MWF"]

    @property
    def T1(self):  # noqa
        return np.stack((self._properties["T1w"], self._properties["T1m"]))

    @property
    def T2(self):  # noqa
        return np.stack((self._properties["T2w"], self._properties["T2m"]))

    @property
    def k(self):  # noqa
        return self._properties["k"]

    @property
    def chemshift(self):  # noqa
        return self._properties["chemshift"]

    @property
    def properties(self):  # noqa
        _properties = {
            "weight": self.MWF,
            "T1": self.T1,
            "T2": self.T2,
            "k": self.k,
            "chemshift": self.chemshift,
        }
        return _properties


class NumericMWSheppLoganPhantom(
    CrispMWSheppLoganPhantom, NumericSheppLoganPhantom
):  # noqa
    """Numeric MW Shepp-Logan phantom builder."""

    pass



"""Base BrainWeb phantom builder class."""

__all__ = ["BrainwebPhantom"]

import os
import numpy as np

from typing import Sequence

from brainweb_dl._brainweb import BIG_RES_SHAPE, BIG_RES_MM

from .._build import PhantomMixin
from .._utils import CacheDirType, get_mrtwin_dir

from ._segmentation import get_brainweb_segmentation


class BrainwebPhantom(PhantomMixin):
    """Base BrainWeb phantom builder."""

    def __init__(
        self,
        ndim: int,
        subject: int,
        shape: int | Sequence[int] | None = None,
        output_res: float | Sequence[float] | None = None,
        cache: bool = True,
        cache_dir: CacheDirType = None,
        brainweb_dir: CacheDirType = None,
        force: bool = False,
        verify: bool = True,
    ):
        # keep dim
        self._ndim = ndim

        # default fov, resolution
        shape, output_res = self._default_prescription(ndim, shape, output_res)

        # get filename
        _fname = self.get_filename(ndim, subject, shape, output_res)

        # try to load segmentation
        self.segmentation, file_path = self.get_segmentation(
            _fname,
            ndim,
            subject,
            shape,
            output_res,
            cache,
            cache_dir,
            brainweb_dir,
            force,
            verify,
        )

        # cache the result
        if cache:
            self.cache(file_path, self.segmentation)

    def _default_prescription(
        self,
        ndim: int,
        output_shape: int | Sequence[int] | None,
        output_res: float | Sequence[float] | None,
    ):
        # original shape and resolution
        orig_shape = np.asarray(BIG_RES_SHAPE)[-ndim:]
        orig_res = np.asarray(BIG_RES_MM)[-ndim:]
        orig_fov = orig_shape * orig_res

        # default shape
        if output_shape is None:
            output_shape = ndim * [200]
        if output_shape is not None and np.isscalar(output_shape):
            output_shape = ndim * [output_shape]
        if output_shape is not None:
            assert len(output_shape) == ndim, ValueError(
                "If shape is not None, it must be either a scalar or a ndim-length sequence."
            )
        output_shape = np.asarray(output_shape)

        # default resolution
        if output_res is not None and np.isscalar(output_res):
            output_res = ndim * [output_res]
        if output_res is not None:
            assert len(output_res) == ndim, ValueError(
                "If output_res is not None, it must be either or a scalar a ndim-length sequence."
            )
        if output_res is None:
            output_res = [max(orig_fov) / max(output_shape)] * ndim
        output_res = np.asarray(output_res)

        return output_shape, output_res

    def __repr__(self):  # noqa
        if self.segmentation is None:
            ptype = "Dense"
        elif len(self.segmentation.shape) == self._ndim:
            ptype = "Crisp"
        else:
            ptype = "Fuzzy"
        msg = f"{ptype} Brainweb phantom with following properties:\n"
        msg += f"Number of spatial dimensions: {self._ndim}\n"
        msg += f"Tissue properties: {self._properties.keys()}\n"
        if self.segmentation is not None:
            _shape = self.shape[-self._ndim :]
        else:
            _shape = list(self._properties.values())[0].shape[-self._ndim :]
        msg += f"Matrix size: {_shape}\n"
        if self.segmentation is not None and len(self.segmentation.shape) != self._ndim:
            msg += f"Number of tissue classes: {self.segmentation.shape[0]}\n"
        return msg

    def get_filename(
        self,
        ndim: int,
        subject: int,
        shape: int | Sequence[int],
        resolution: float | Sequence[float],
    ):
        """
        Generate filename starting from FOV and matrix shape.

        Parameters
        ----------
        ndim : int
            Number of spatial dimensions. If ndim == 2, use a single slice
            (central axial slice).
        shape: int | Sequence[int]
            Shape of the output data, the data will be interpolated to the given shape.
            If int, assume isotropic matrix.
        resolution: float | Sequence[float]
            Resolution of the output data, the data will be rescale to the given resolution.
            If scalar, assume isotropic resolution.

        Returns
        -------
        str
            Filename for caching.

        """
        _fov = np.ceil(shape * resolution).astype(int).tolist()

        fov_str = [str(ri) for ri in _fov]
        shape_str = [str(di) for di in shape.tolist()]

        fov_str = "x".join(tuple(fov_str))
        shape_str = "x".join(tuple(shape_str))

        return f"{self.__class__.__name__.lower()}{subject:02d}_{fov_str}fov_{shape_str}mtx.npy"

    def get_segmentation(
        self,
        fname: str,
        ndim: int,
        subject: int,
        shape: int | Sequence[int],
        output_res: float | Sequence[float],
        cache: bool,
        cache_dir: CacheDirType,
        brainweb_dir: CacheDirType,
        force: bool,
        verify: bool,
    ):
        """
        Get fuzzy BrainWeb tissue segmentation.

        Parameters
        ----------
        fname : str
            Filename for caching.
        ndim : int
            Number of spatial dimensions. If ndim == 2, use a single slice
            (central axial slice).
        subject : int
            Subject id to download.
        shape: int | Sequence[int]
            Shape of the output data, the data will be interpolated to the given shape.
            If int, assume isotropic matrix.
        output_res: float | Sequence[float]
            Resolution of the output data, the data will be rescale to the given resolution.
            If scalar, assume isotropic resolution.
        cache : bool
            If True, cache the result.
        cache_dir : CacheDirType
            Directory for segmentation caching.
        brainweb_dir : CacheDirType
            Brainweb_directory to download the data.
        force : bool
            Force download even if the file already exists.
        verify : bool
            Enable SSL verification.
            DO NOT DISABLE (i.e., verify=False)IN PRODUCTION.

        Returns
        -------
        np.ndarray.
            Brainweb segmentation.
        file_path : str
            Path on disk to generated segmentation for caching.

        """
        # get base directory
        cache_dir = get_mrtwin_dir(cache_dir)

        # get file path
        file_path = os.path.join(cache_dir, fname)

        # try to load
        if os.path.exists(file_path) and not (force):
            return np.load(file_path), file_path
        else:
            segmentation = get_brainweb_segmentation(
                ndim, subject, shape, output_res, brainweb_dir, force, verify
            )

        return segmentation, file_path

    def __array__(self):  # noqa
        # This method tells NumPy how to convert the object to an array
        return self.segmentation



"""Miscellaneous routines sub-package."""

__all__ = []

from ._rigid_motion import rigid_motion  # noqa
from ._girf import generate_girf  # noqa

__all__.append("rigid_motion")
__all__.append("generate_girf")



"""Test rigid motion pattern generation."""

import pytest
import numpy as np

from mrtwin import rigid_motion


@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("nframes", [10, 100])
@pytest.mark.parametrize("degree", ["subtle", "moderate", "severe"])
def test_rigid_motion_basic(ndim, nframes, degree):
    """
    Test the rigid motion generator function for basic cases with predefined degrees of motion.
    """
    # Call the rigid motion function
    params = rigid_motion(ndim, nframes, degree=degree)

    if ndim == 2:
        angles = tuple([params[0]])
        translations = params[1:]
    elif ndim == 3:
        angles = params[:3]
        translations = params[3:]

    # Validate the shape of the returned arrays
    for angle in angles:
        assert angle.shape == (
            nframes,
        ), f"Expected angle shape to be {(nframes,)}, but got {angle.shape}."
    for translation in translations:
        assert translation.shape == (
            nframes,
        ), f"Expected translation shape to be {(nframes,)}, but got {translation.shape}."


@pytest.mark.parametrize(
    "degree, max_rot, max_trans",
    [
        ("subtle", 5.0, 2.0),
        ("moderate", 10.0, 8.0),
        ("severe", 16.0, 16.0),
    ],
)
def test_rigid_motion_degree_values(degree, max_rot, max_trans):
    """
    Test the rigid motion generator with predefined severity levels to ensure the values fall within expected ranges.
    """
    ndim = 3
    nframes = 50

    params = rigid_motion(ndim, nframes, degree=degree)
    angles = params[:3]
    translations = params[3:]

    # Check that rotation values are within the expected range
    for angle in angles:
        assert np.all(
            np.abs(angle) <= max_rot
        ), f"Expected max rotation {max_rot} degrees, but got a value exceeding that."

    # Check that translation values are within the expected range
    for translation in translations:
        assert np.all(
            np.abs(translation) <= max_trans
        ), f"Expected max translation {max_trans} mm, but got a value exceeding that."


@pytest.mark.parametrize("degree", [(7.0, 3.5), (15.0, 6.0)])
def test_rigid_motion_custom_degree(degree):
    """
    Test the rigid motion generator with custom degrees for both rotation and translation.
    """
    ndim = 3
    nframes = 50

    params = rigid_motion(ndim, nframes, degree=degree)
    angles = params[:3]
    translations = params[3:]

    # Validate the max rotation and translation
    max_rot, max_trans = degree

    for angle in angles:
        assert np.all(
            np.abs(angle) <= max_rot
        ), f"Expected max rotation {max_rot} degrees, but got a value exceeding that."
    for translation in translations:
        assert np.all(
            np.abs(translation) <= max_trans
        ), f"Expected max translation {max_trans} mm, but got a value exceeding that."


def test_rigid_motion_seed():
    """
    Test that the random seed produces reproducible results.
    """
    ndim = 3
    nframes = 50
    degree = "moderate"
    seed = 42

    # Generate motion with the seed
    params1 = rigid_motion(ndim, nframes, degree=degree, seed=seed)
    angles1 = params1[:3]
    translations1 = params1[3:]

    params2 = rigid_motion(ndim, nframes, degree=degree, seed=seed)
    angles2 = params1[:3]
    translations2 = params2[3:]

    # Ensure that results are identical when the seed is set
    for a1, a2 in zip(angles1, angles2):
        np.testing.assert_array_equal(
            a1, a2, err_msg="Angles do not match for the same seed."
        )
    for t1, t2 in zip(translations1, translations2):
        np.testing.assert_array_equal(
            t1, t2, err_msg="Translations do not match for the same seed."
        )


@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("nframes", [10, 100])
def test_rigid_motion_dimensions(ndim, nframes):
    """
    Test that the number of dimensions (2D/3D) affects the number of outputs correctly.
    """
    degree = "moderate"

    params = rigid_motion(ndim, nframes, degree=degree)
    angles = params[:3]
    translations = params[3:]

    if ndim == 2:
        angles = tuple([params[0]])
        translations = params[1:]
    elif ndim == 3:
        angles = params[:3]
        translations = params[3:]

    # Validate that the correct number of angles and translations are generated based on ndims
    if ndim == 2:
        assert len(angles) == 1, f"Expected 1 angle, but got {len(angles)}."
    elif ndim == 3:
        assert len(angles) == ndim, f"Expected {ndim} angles, but got {len(angles)}."
    assert (
        len(translations) == ndim
    ), f"Expected {ndim} translations, but got {len(translations)}."



"""Gradient Impulse Response Function generation routines."""

__all__ = ["generate_girf"]

from typing import Sequence


import numpy as np


def generate_girf(
    dt: float, N: int, fwhm: float | Sequence[float], delay: float | Sequence[float]
):
    """
    Generate an approximate Gradient Impulse Response Function (GIRF).

    The GIRF is approximated as a complex-valued array, whose magnitude (modeled)
    as a Gaussian of the specified Full Width Half Maximum, which dampens
    high-frequency components of the k-space trajectory, and whose phase
    is a linear ramp representing a non-integer shift of the waveform.

    The response is modeled independently for the three physical axes ``z, y, x``.

    Parameters
    ----------
    dt : float
        Gradient raster time (in seconds).
    N : int
        Length of the GIRF.
    fwhm : float | Sequence[float]
        Width in Hz of the magnitude response of the gradient
        system for the three axes ``z, y, x``.
        If it is a scalar, assume the
        same delay for all the axes.
    delay : float | Sequence[float]
        Non-integer delay (in seconds) for the three axes
        ``z, y, x``. If it is a scalar, assume the
        same delay for all the axes.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        A tuple containing:
        - The frequency axis (in Hz) as a NumPy array.
        - Complex GIRF of shape ``(3, N)``.

    Notes
    -----
    Typical values for gradient raster times ``dt`` are around 5-10 us
    (GE: 4us; Siemens: 10us; Philips: 6.4us)

    GIRF magnitudes Full Width Half Maximum are on the order of 5 kHz,
    while gradient delays are on the order of 1us.

    GIRF for the x and y axes are more similar to each other
    when compared to z axes.

    GIRF is typically measured (via thin slice method or dynamic field camera)
    with a time window T around 100ms (N = T / dt).

    Examples
    --------
    >>> from mrtwin import generate_girf

    Assuming same fwhm and delay for each axes, we have:

    >>> freqs, girf = generate_girf(4e-6, 25000, 5.0e3, 1.1e-6)

    """
    # Handle scalar case
    if np.isscalar(fwhm):
        fwhm = [fwhm] * 3
    fwhm = list(fwhm)
    if np.isscalar(delay):
        delay = [delay] * 3
    delay = list(delay)

    # Calculate frequency axis (Hz)
    freqs = np.fft.fftfreq(N, d=dt)

    # Generate GIRF magnitude
    girf_magn = np.stack([_gaussian_freq_domain(dt, N, f) for f in fwhm], axis=0)
    girf_magn = girf_magn.astype(np.float32)

    # Generate GIRF phase
    girf_phase = np.stack([_phase_shift_line(dt, N, d) for d in delay], axis=0)
    girf_phase = girf_phase.astype(np.float32)

    return freqs, girf_magn * np.exp(1j * girf_phase)


def _gaussian_freq_domain(dt: float, N: int, fwhm: float):
    # Calculate frequency axis (Hz)
    freqs = np.fft.fftfreq(N, d=dt)

    # Calculate standard deviation from FWHM
    sigma_f = fwhm / (2 * np.sqrt(2 * np.log(2)))

    # Generate Gaussian in the frequency domain
    gaussian = np.exp(-0.5 * (freqs / sigma_f) ** 2)

    return gaussian


def _phase_shift_line(dt: float, N: int, delay: float):
    # Calculate frequency axis (Hz)
    freqs = np.fft.fftfreq(N, d=dt)

    # Generate the linear phase ramp: -2 * pi * f * delay
    phase_ramp = -2 * np.pi * freqs * delay

    return phase_ramp



"""Test sensitivity maps generation."""

import pytest
import numpy as np

from mrtwin import sensmap


@pytest.mark.parametrize("shape", [(8, 128, 128), (8, 64, 64), (8, 64, 64, 64)])
def test_sensmap_basic_shape(shape):
    """
    Test the sensmap function with basic shape parameters.
    """
    smap = sensmap(shape)

    # Validate the output is not None
    assert smap is not None, "Sensmap output should not be None."

    # Validate the output type is numpy.ndarray
    assert isinstance(smap, np.ndarray), "Sensmap output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert smap.shape == shape, f"Expected shape {shape}, but got {smap.shape}."

    # Check that the output contains complex numbers
    assert np.iscomplexobj(smap), "Sensmap output should be a complex numpy array."


@pytest.mark.parametrize("coil_width", [1.0, 2.0, 3.5])
def test_sensmap_coil_width(coil_width):
    """
    Test the sensmap function with different coil_width values.
    """
    shape = (8, 128, 128)
    smap = sensmap(shape, coil_width=coil_width)

    # Validate the output is a numpy ndarray
    assert isinstance(smap, np.ndarray), "Sensmap output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert smap.shape == shape, f"Expected shape {shape}, but got {smap.shape}."


@pytest.mark.parametrize("shift", [None, (0, 0), (5, -5), (10, 10, 10)])
def test_sensmap_shift(shift):
    """
    Test the sensmap function with different shift values.
    """
    shape = (8, 128, 128)
    if len(shift or ()) == 3:
        shape = (8, 64, 64, 64)
    smap = sensmap(shape, shift=shift)

    # Validate the output is a numpy ndarray
    assert isinstance(smap, np.ndarray), "Sensmap output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert smap.shape == shape, f"Expected shape {shape}, but got {smap.shape}."


@pytest.mark.parametrize("dphi", [0.0, 15.0, 90.0])
def test_sensmap_dphi(dphi):
    """
    Test the sensmap function with different dphi values (coil rotation angle).
    """
    shape = (8, 128, 128)
    smap = sensmap(shape, dphi=dphi)

    # Validate the output is a numpy ndarray
    assert isinstance(smap, np.ndarray), "Sensmap output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert smap.shape == shape, f"Expected shape {shape}, but got {smap.shape}."


@pytest.mark.parametrize("nrings", [None, 2, 4])
def test_sensmap_nrings(nrings):
    """
    Test the sensmap function with different nrings values.
    """
    shape = (8, 128, 128)
    smap = sensmap(shape, nrings=nrings)

    # Validate the output is a numpy ndarray
    assert isinstance(smap, np.ndarray), "Sensmap output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert smap.shape == shape, f"Expected shape {shape}, but got {smap.shape}."


def test_sensmap_default():
    """
    Test the sensmap function with default parameters.
    """
    shape = (8, 128, 128)
    smap = sensmap(shape)

    # Validate the output is a numpy ndarray
    assert isinstance(smap, np.ndarray), "Sensmap output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert smap.shape == shape, f"Expected shape {shape}, but got {smap.shape}."

    # Validate the output contains complex numbers
    assert np.iscomplexobj(smap), "Sensmap output should be a complex numpy array."



"""Set spatial prescription for a given phantom."""

__all__ = ["set_prescription"]

from typing import Sequence

import numpy as np

from . import _utils


def set_prescription(
    data: np.ndarray,
    orig_res: Sequence[float],
    orig_shape: Sequence[int],
    output_res: Sequence[float],
    output_shape: Sequence[int] | None = None,
):
    """
    Set prescription (fov and resolution) for an input dataset.

    Parameters
    ----------
    data : np.ndarray
        Input dataset of shape (..., (nz0), ny0, nx0).
    orig_res : Sequence[float]
        Input resolution ((dz0), dy0, dx0), same units as output_res.
    orig_shape : Sequence[int]
        Input shape ((nz0), ny0, nx0).
    output_res : Sequence[float]
        Output resolution ((dz1), dy1, dx1), same units as orig_res.
    output_shape : Sequence[int] | None, optional
        Output shape ((nz1), ny1, nx1). If not provided,
        calculate shape to preserve original FoV
        ((dz0 * nz0), dy0 * ny0, dx0 * nx0).

    Returns
    -------
    data : np.ndarray
        Resampled data to ((nz1), ny1, nx1) so that
        output resolution is ((dz1), dy1, dx1).

    """
    # c0nvert to array
    orig_res = np.asarray(orig_res)
    output_res = np.asarray(output_res)

    # get fov
    orig_fov = np.asarray(orig_shape) * orig_res

    # default output shape
    if output_shape is None:
        output_shape = orig_fov / output_res
        output_shape = np.ceil(output_shape).astype(int)

    # get output fov
    output_fov = output_shape * output_res

    # select region of interest
    scale = output_fov / orig_fov
    roi = scale * np.asarray(orig_shape)
    roi = np.round(roi).astype(int)

    # crop or pad
    data = _utils.resize(data, roi)

    # resample to desired output resolution
    data = _utils.resample(data, output_shape)

    return data



"""
===========================================
Gradien System Response Function Generation
===========================================

Example of Gradien System Response Function (GIRF) generation.

The GIRF is approximated as a complex-valued array, whose magnitude (modeled)
as a Gaussian of the specified Full Width Half Maximum, which dampens
high-frequency components of the k-space trajectory, and whose phase
is a linear ramp representing a non-integer shift of the waveform.

The response is modeled independently for the three physical axes ``z, y, x``.

"""

import matplotlib.pyplot as plt
import numpy as np

from mrtwin import generate_girf

# %%
#
# Typical values for gradient raster times ``dt`` are around 5-10 us
# (GE: 4us; Siemens: 10us; Philips: 6.4us)
#
# GIRF magnitudes Full Width Half Maximum are on the order of 5 kHz,
# while gradient delays are on the order of 1us.
#
# GIRF for the x and y axes are more similar to each other
# when compared to z axes.
#
# Given a GIRF measurement time window of 100ms, function parameters are
# as follows:

dt = 4e-6  # assume GE raster time
N = int(100e-3 // dt)  # approximate number of points
fwhm = (8.0e3, 5.9e3, 6.1e3)
delay = (3.2e-6, 0.9e-6, 1.1e-6)

# %%
# Given these parameters, GIRF can be computed as

freqs, girf = generate_girf(dt, N, fwhm, delay)

fig, ax = plt.subplots(2, 1)
ax[0].plot(freqs * 1e-3, np.abs(girf).T, "."), ax[0].set_xlim([-10, 10]), ax[0].legend(
    ["$GIRF_z$", "$GIRF_y$", "$GIRF_x$"]
), ax[0].set(xlabel="frequency [kHz]", ylabel="GIRF magnitude")
ax[1].plot(freqs * 1e-3, np.angle(girf).T, "."), ax[1].set_xlim([-10, 10]), ax[
    1
].set_ylim([-0.5, 0.5]), ax[1].set(xlabel="frequency [kHz]", ylabel="GIRF phase [rad]")



"""Array shape manipulation routines."""

__all__ = ["resize"]

import numpy as np


from ._broadcasting import _expand_shapes


def resize(input, oshape):
    """
    Resize with zero-padding or cropping.

    Adapted from SigPy [1].

    Parameters
    ----------
    input : np.ndarray
        Input array of shape ``(..., ishape)``.
    oshape : Sequence
        Output shape.

    Returns
    -------
    output : np.ndarray
        Zero-padded or cropped array of shape ``(..., oshape)``.

    Examples
    --------
    >>> import numpy as np
    >>> import mrtwin

    We can pad tensors to desired shape:

    >>> x = np.asarray([0, 1, 0])
    >>> y = mrtwin.resize(x, [5])
    >>> y
    array([0, 0, 1, 0, 0])

    Batch dimensions are automatically expanded (pad will be applied starting from rightmost dimension):

    >>> x = np.asarray([0, 1, 0])[None, ...]
    >>> x.shape
    (1, 3)
    >>> y = mrtwin.resize(x, [5]) # len(oshape) == 1
    >>> y.shape
    (1, 5)

    Similarly, if oshape is smaller than ishape, the tensor will be cropped:

    >>> x = np.asarray([0, 0, 1, 0, 0])
    >>> y = mrtwin.resize(x, [3])
    >>> y
    tensor([0, 1, 0])

    Again, batch dimensions are automatically expanded:

    >>> x = np.asarray([0, 0, 1, 0, 0])[None, ...]
    >>> x.shape
    (1, 5)
    >>> y = mrtwin.resize(x, [3]) # len(oshape) == 1
    >>> y.shape
    (1, 3)

    References
    ----------
    [1] https://github.com/mikgroup/sigpy/blob/main/sigpy/util.py

    """
    if isinstance(oshape, int):
        oshape = [oshape]

    ishape1, oshape1 = _expand_shapes(input.shape, oshape)

    if ishape1 == oshape1:
        return input

    # shift not supported for now
    ishift = [max(i // 2 - o // 2, 0) for i, o in zip(ishape1, oshape1)]
    oshift = [max(o // 2 - i // 2, 0) for i, o in zip(ishape1, oshape1)]

    copy_shape = [
        min(i - si, o - so) for i, si, o, so in zip(ishape1, ishift, oshape1, oshift)
    ]
    islice = tuple([slice(si, si + c) for si, c in zip(ishift, copy_shape)])
    oslice = tuple([slice(so, so + c) for so, c in zip(oshift, copy_shape)])

    output = np.zeros(oshape1, dtype=input.dtype)
    input = input.reshape(ishape1)
    output[oslice] = input[islice]

    return output



"""
===============================
Rigid Motion Pattern Generation
===============================

Example of rigid motion pattern generation.

This examples show how to generate rigid motion pattern
for 2D (in-plane rotation, x-y translation) and 3D (3D rotation, x-y-z translation)
imaging problems.
"""

import matplotlib.pyplot as plt
import numpy as np

from mrtwin import rigid_motion

# %%
# Basic Usage
# ===========
#
# Two- and three-dimensional rigid motion patterns
# can be generated as follows, using ``nframes`` to
# specify the number of motion states:

theta, dy2D, dx2D = rigid_motion(2, 200 * 200)
roll, pitch, yaw, dz3D, dy3D, dx3D = rigid_motion(3, 200 * 200)

fig1, ax1 = plt.subplots(2, 2)
ax1[0, 0].plot(theta, "."), ax1[0, 0].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax1[0, 0].set_title("2D motion")
ax1[1, 0].plot(np.stack((dy2D, dx2D), axis=1), "."), ax1[1, 0].set(
    xlabel="motion state #", ylabel="translation [mm]"
), ax1[1, 0].set_title("2D motion")
ax1[0, 1].plot(np.stack((roll, pitch, yaw), axis=1), "."), ax1[0, 1].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax1[0, 1].set_title("3D motion")
ax1[1, 1].plot(np.stack((dz3D, dy3D, dx3D), axis=1), "."), ax1[1, 1].set(
    xlabel="motion state #", ylabel="translation [mm]"
), ax1[1, 1].set_title("23D motion")
plt.show()

# %%
# For reproducibility, the seed for random pattern generation is set to ``42``.
# This can be changed via the ``seed`` argument:

theta1, dy1, dx1 = rigid_motion(2, 200 * 200, seed=10)
theta2, dy2, dx2 = rigid_motion(2, 200 * 200, seed=20)
theta3, dy3, dx3 = rigid_motion(2, 200 * 200, seed=30)

fig2, ax2 = plt.subplots(2, 3)
ax2[0, 0].plot(theta1, "."), ax2[0, 0].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax2[0, 0].set_title("seed = 10")
ax2[1, 0].plot(np.stack((dy1, dx1), axis=1), "."), ax2[1, 0].set(
    xlabel="motion state #", ylabel="translation [mm]"
)
ax2[0, 1].plot(theta2, "."), ax2[0, 1].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax2[0, 1].set_title("seed = 20")
ax2[1, 1].plot(np.stack((dy2, dx2), axis=1), "."), ax2[1, 1].set(
    xlabel="motion state #", ylabel="translation [mm]"
)
ax2[0, 2].plot(theta3, "."), ax2[0, 2].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax2[0, 2].set_title("seed = 20")
ax2[1, 2].plot(np.stack((dy3, dx3), axis=1), "."), ax2[1, 2].set(
    xlabel="motion state #", ylabel="translation [mm]"
)
plt.show()

# %%
# Advanced Usage
# ==============
#
# Severity of motion can be specified via the ``degree`` argument.
# This can be a string - accepted values are ``"subtle"``, ``"moderate"``
# and ``"severe"``. These corresponds to the following motion ranges:

# * ``"subtle"``: maximum rotation ``5.0 [deg]``; maximum translation ``2.0 [mm]``
# * ``"moderate"``: maximum rotation ``10.0 [deg]``; maximum translation ``8.0 [mm]``
# * ``"severe"``: maximum rotation ``16.0 [deg]``; maximum translation ``16.0 [mm]`

theta1, dy1, dx1 = rigid_motion(2, 200 * 200, degree="subtle")
theta2, dy2, dx2 = rigid_motion(2, 200 * 200, degree="moderate")
theta3, dy3, dx3 = rigid_motion(2, 200 * 200, degree="severe")

fig3, ax3 = plt.subplots(2, 3)
ax3[0, 0].plot(theta1, "."), ax3[0, 0].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax3[0, 0].set_title("subtle motion")
ax3[1, 0].plot(np.stack((dy1, dx1), axis=1), "."), ax3[1, 0].set(
    xlabel="motion state #", ylabel="translation [mm]"
)
ax3[0, 1].plot(theta2, "."), ax3[0, 1].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax3[0, 1].set_title("moderate motion")
ax3[1, 1].plot(np.stack((dy2, dx2), axis=1), "."), ax3[1, 1].set(
    xlabel="motion state #", ylabel="translation [mm]"
)
ax3[0, 2].plot(theta3, "."), ax3[0, 2].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax3[0, 2].set_title("severe motion")
ax3[1, 2].plot(np.stack((dy3, dx3), axis=1), "."), ax3[1, 2].set(
    xlabel="motion state #", ylabel="translation [mm]"
)
plt.show()

# %%
#
# As an alternative, user can specify a tuple of floats, where ``degree[0]``
# is the maximum rotation in ``[deg]`` and ``degree[1]`` is the maximum translation
# in ``[mm]``:

theta1, dy1, dx1 = rigid_motion(2, 200 * 200, degree=(30, 5))

fig4, ax4 = plt.subplots(1, 2)
ax4[0].plot(theta1, "."), ax4[0].set(
    xlabel="motion state #", ylabel="rotation angle [deg]"
), ax4[0].set_title("custom motion")
ax4[1].plot(np.stack((dy1, dx1), axis=1), "."), ax4[1].set(
    xlabel="motion state #", ylabel="translation [mm]"
)
plt.show()



"""Wrapper around Phantominator to extract MR tissue segmentation."""

__all__ = ["get_shepp_logan"]

import importlib

from typing import Dict, Sequence

import numpy as np
import numpy.typing as npt

mr_shepp_logan = importlib.import_module("phantominator.mr_shepp_logan")


def _mr_relaxation_parameters() -> Dict[str, npt.ArrayLike]:
    """Return MR relaxation parameters for certain tissues.

    Returns
    -------
    params : dict
        Gives entries as [A, C, (t1), t2, chi]

    Notes
    -----
    If t1 is None, the model T1 = A*B0^C will be used.  If t1 is not
    np.nan, then specified t1 will be used.
    """
    # params['tissue-name'] = [A, C, (t1 value if explicit), t2, chi]
    params = dict()
    params["scalp"] = [np.nan, np.nan, 9, 0.07, -7.5e-6]
    params["marrow"] = [np.nan, np.nan, 11, 0.05, -8.85e-6]
    params["csf"] = [np.nan, np.nan, 1, 1.99, -9e-6]
    params["blood-clot"] = [np.nan, np.nan, 8, 0.2, -9e-6]
    params["gray-matter"] = [np.nan, np.nan, 2, 0.1, -9e-6]
    params["white-matter"] = [np.nan, np.nan, 3, 0.08, -9e-6]
    params["tumor"] = [np.nan, np.nan, 8, 0.1, -9e-6]
    return params


def mr_ellipsoid_parameters() -> npt.ArrayLike:
    """Return parameters of ellipsoids.

    Returns
    -------
    E : array_like
        Parameters for the ellipsoids used to construct the phantom.
    """
    params = _mr_relaxation_parameters()

    E = np.zeros((15, 13))
    # [:, [x, y, z, a, b, c, theta, m0, A, C, (t1), t2, chi]]
    E[0, :] = [0, 0, 0, 0.72, 0.95, 0.93, 0, 0.8, *params["scalp"]]
    E[1, :] = [0, 0, 0, 0.69, 0.92, 0.9, 0, 0.12, *params["marrow"]]
    E[2, :] = [0, -0.0184, 0, 0.6624, 0.874, 0.88, 0, 0.98, *params["gray-matter"]]
    E[3, :] = [0, -0.0184, 0, 0.6024, 0.814, 0.82, 0, 0.745, *params["white-matter"]]
    E[4, :] = [-0.3, 0, 0, 0.41, 0.16, 0.21, np.deg2rad(-72), 0.98, *params["csf"]]
    E[5, :] = [0.25, 0, 0, 0.31, 0.11, 0.22, np.deg2rad(72), 0.98, *params["csf"]]
    E[6, :] = [0, 0.35, 0, 0.15, 0.15, 0.15, 0, 0.617, *params["tumor"]]
    E[7, :] = [0, 0.1, 0, 0.046, 0.046, 0.046, 0, 0.95, *params["tumor"]]
    E[8, :] = [-0.08, -0.605, 0, 0.046, 0.023, 0.02, 0, 0.95, *params["tumor"]]
    E[9, :] = [
        0.06,
        -0.605,
        0,
        0.046,
        0.023,
        0.02,
        np.deg2rad(-90),
        0.95,
        *params["tumor"],
    ]
    E[10, :] = [0, -0.1, 0, 0.046, 0.046, 0.046, 0, 0.95, *params["tumor"]]
    E[11, :] = [0, -0.605, 0, 0.023, 0.023, 0.023, 0, 0.95, *params["tumor"]]
    E[12, :] = [
        0.06,
        -0.105,
        0.3125,
        0.056,
        0.04,
        0.1,
        np.deg2rad(-90),
        0.93,
        *params["tumor"],
    ]
    E[13, :] = [0, 0.1, 0.3125, 0.056, 0.056, 0.1, 0, 0.98, *params["csf"]]
    E[14, :] = [
        0.56,
        -0.4,
        0.25,
        0.2,
        0.03,
        0.1,
        np.deg2rad(70),
        0.85,
        *params["blood-clot"],
    ]

    # Need to subtract some ellipses here...
    Eneg = np.zeros(E.shape)
    for ii in range(E.shape[0]):

        # Ellipsoid geometry
        Eneg[ii, :7] = E[ii, :7]

        # Tissue property differs after 4th subtracted ellipsoid
        if ii > 3:
            Eneg[ii, 7:] = E[3, 7:]
        else:
            Eneg[ii, 7:] = E[ii - 1, 7:]

    # Throw out first as we skip this one in the paper's table
    Eneg = Eneg[1:, :]

    # Spin density is negative for subtraction
    Eneg[:, 7] *= -1

    # Paper doesn't use last blood-clot ellipsoid
    E = E[:-1, :]
    Eneg = Eneg[:-1, :]

    # Put both ellipsoid groups together
    E = np.concatenate((E, Eneg), axis=0)

    return E


# Monkey patch
mr_shepp_logan.mr_ellipsoid_parameters = mr_ellipsoid_parameters


def get_shepp_logan(ndim: int, shape: int | Sequence[int]):
    """
    Get crisp Shepp-Logan tissue segmentation.

    Parameters
    ----------
    ndim : int
        Number of spatial dimensions. If ndim == 2, create a single slice
        2D phantom.
    shape: int | Sequence[int]
        Shape of the output data, the data will be interpolated to the given shape.
        If int, assume isotropic matrix.

    Returns
    -------
    np.ndarray.
        Shepp-Logan segmentation.

    """
    assert ndim == 2 or ndim == 3, ValueError(
        f"Number of spatial dimensions (={ndim}) must be either 2 or 3."
    )

    # default params
    if np.isscalar(shape):
        shape = shape * np.ones(ndim, dtype=int)
    elif ndim == 3:
        shape = [shape[1], shape[2], shape[0]]

    # cast to list
    shape = list(shape)
    assert len(shape) == ndim, ValueError(
        "If shape must be either a scalar or a ndim-length sequence."
    )

    if ndim == 2:
        shape = shape + [5]
        zlims = (0, 0.5)
    else:
        zlims = (-1, 1)

    # build segmentation
    _, data, _ = mr_shepp_logan.mr_shepp_logan(shape, zlims=zlims)

    if ndim == 2:
        data = data[:, :, 0]
    else:
        data = data.transpose(-1, 0, 1)

    return np.ascontiguousarray(data).astype(int)



"""Centered ND FFT subroutines."""

__all__ = ["fftc", "ifftc"]

import numpy as np


def fftc(x, ax):  # noqa
    return np.fft.fftshift(
        np.fft.fftn(np.fft.ifftshift(x, axes=ax), axes=ax, norm=None), axes=ax
    )


def ifftc(x, ax):  # noqa
    return np.fft.fftshift(
        np.fft.ifftn(np.fft.ifftshift(x, axes=ax), axes=ax, norm=None), axes=ax
    )



"""Coil sensitivity maps generation routines."""

__all__ = ["sensmap"]


import os


from typing import Sequence


import numpy as np


from .._utils import CacheDirType, get_mrtwin_dir


from ._birdcage import _birdcage


def sensmap(
    shape: Sequence[int],
    coil_width: float = 2.0,
    shift: Sequence[int] | None = None,
    dphi: float = 0.0,
    nrings: int = None,
    cache: bool | None = None,
    cache_dir: CacheDirType = None,
):
    """
    Simulate birdcage coils.

    Adapted from SigPy [1].

    Parameters
    ----------
    shape : Iterable[int]
        Size of the matrix ``(ncoils, ny, nx)`` (2D)
        or ``(ncoils, nz, ny, nx)`` (3D) for the sensitivity coils.
    coil_width : float, optional
        Width of the coil, with respect to image dimension.
        The default is ``2.0``.
    shift : Sequence[int] | None, optional
        Displacement of the coil center with respect to matrix center.
        The default is ``(0, 0)`` / ``(0, 0, 0)``.
    dphi : float, optional
        Bulk coil angle in ``[deg]``.
        The default is ``0.0°``.
    nrings : int | None, optional
        Number of rings for a cylindrical hardware set-up.
        The default is ``ncoils // 4``.
    cache : bool | None, optional
        If ``True``, cache the phantom.
        The default is ``True`` for 3D phantoms
        and ``False`` for single-slice 2D.
    cache_dir : CacheDirType, optional
        cache_directory for phantom caching.
        The default is ``None`` (``~/.cache/mrtwin``).

    Returns
    -------
    smap : torch.Tensor
        Complex spatially varying sensitivity maps of shape ``(nmodes, ny, nx)`` (2D)
        or ``(nmodes, nz, ny, nx)`` (3D). If ``nmodes = 1``, the first dimension is squeezed.

    Example
    -------
    >>> from mrtwin import sensmap

    We can generate a set of ``nchannels=8`` 2D sensitivity maps of shape ``(ny=128, nx=128)`` by:

    >>> smap = sensmap((8, 128, 128))

    Coil center and rotation can be modified by ``shift`` and ``dphi`` arguments:

    >>> smap = sensmap((8, 128, 128), shift=(-3, 5), dphi=30.0) # center shifted by (dy, dx) = (-3, 5) pixels and rotated by 30.0 degrees.

    Similarly, ``nchannels=8`` 3D sensitivity maps can be generated as:

    >>> smap = sensmap((8, 128, 128, 128))

    Beware that this will require more memory.

    References
    ----------
    [1] https://github.com/mikgroup/sigpy/tree/main

    """
    # Default values
    if shift is None:
        shift = [0.0 for ax in range(len(shape) - 1)]
    if nrings is None:
        nrings = np.max((shape[0] // 4, 1))
    if cache is None and len(shape) == 3:  # (nc, ny, nx) -> 2D
        cache = False
    elif cache is None and len(shape) == 4:  # (nc, nz, ny, nx) -> 3D
        cache = True

    # Get filename for caching
    shape_str = [str(di) for di in shape]
    shape_str = "x".join(tuple(shape_str))

    shift_str = [str(di) for di in shift]
    shift_str = "x".join(tuple(shift_str))

    file_name = f"sensmap{shape_str}mtx_{coil_width}width_{shift_str}shift_{nrings}rings_{dphi}deg.npy"

    # Get base directory
    cache_dir = get_mrtwin_dir(cache_dir)

    # Get file path
    file_path = os.path.join(cache_dir, file_name)

    # Try to load
    if os.path.exists(file_path):
        return np.load(file_path)

    # Generate map
    smap = _birdcage(shape, coil_width, nrings, shift, np.deg2rad(dphi))

    # Normalize
    rss = sum(abs(smap) ** 2, 0) ** 0.5
    smap /= rss

    # Cache the result
    if cache and os.path.exists(file_path) is False:
        np.save(file_path, smap)

    return smap



"""Custom typing for type hints."""

__all__ = ["CacheDirType", "PhantomType"]

from pathlib import Path
from typing import Any, Union

CacheDirType = Union[Path, None]
PhantomType = Any  # TODO: find something more useful



"""Main MR-Twin API."""

__all__ = []

from ._brainweb import brainweb_phantom  # noqa
from ._osf import osf_phantom  # noqa
from ._shepplogan import shepplogan_phantom  # noqa

from ._fieldmap import b0field  # noqa
from ._fieldmap import b1field  # noqa
from ._fieldmap import sensmap  # noqa

from ._misc import rigid_motion  # noqa
from ._misc import generate_girf  # noqa

# Phantoms
__all__.append("brainweb_phantom")
__all__.append("osf_phantom")
__all__.append("shepplogan_phantom")

# Fields
__all__.append("b0field")
__all__.append("b1field")
__all__.append("sensmap")

# Miscellaneous
__all__.append("rigid_motion")
__all__.append("generate_girf")



"""B0 field maps generation routines."""

__all__ = ["b0field"]


from typing import Sequence


import numpy as np


from .._utils import fftc, ifftc


def b0field(
    chi: np.ndarray,
    b0range: Sequence[float] | None = None,
    mask: np.ndarray | None = None,
    B0: float = 1.5,
):
    """
    Simulate inhomogeneous B0 fields.

    Output field units is ``[Hz]``. The field
    is created by convolving the dipole kernel with an input
    magnetic susceptibility map.

    Parameters
    ----------
    chi : np.ndarray
        Object magnetic susceptibility map in ``[ppb]`` of
        shape ``(ny, nx)`` (2D) or ``(nz, ny, nx)`` (3D).
    b0range : Sequence[float] | None, optional
        Range of B0 field in ``[Hz]``. The default is ``None``
        (do not force a range).
    mask : np.ndarray, optional
        Region of support of the object of
        shape ``(ny, nx)`` (2D) or ``(nz, ny, nx)`` (3D).
        The default is ``None``.
    B0 : float, optional
        Static field strength in [T]. Used to compute
        B0 scaling (assuming 1H imaging)
        if `b0range` is not provided.
        The default is `1.5`.

    Returns
    -------
    B0map : np.ndarray
        Spatially varying B0 maps of shape ``(ny, nx)`` (2D)
        or ``(nz, ny, nx)`` (3D) in ``[Hz]``, arising from the object susceptibility.

    Example
    -------
    >>> from mrtwin import shepplogan_phantom, b0field

    We can generate a 2D B0 field map of shape ``(ny=128, nx=128)`` starting from a
    magnetic susceptibility distribution:

    >>> chi = shepplogan_phantom(2, 128, segtype=False).Chi
    >>> b0map = b0field(chi)

    B0 values range can be specified using ``b0range`` argument:

    >>> b0map = b0field(chi, b0range=(-500, 500))

    """
    # get input shape
    ishape = chi.shape

    # get number of spatial dims
    ndim = len(ishape)

    # get k space coordinates
    kgrid = [
        np.arange(-ishape[n] // 2, ishape[n] // 2, dtype=np.float32)
        for n in range(len(ishape))
    ]
    kgrid = np.meshgrid(*kgrid, indexing="ij")
    kgrid = np.stack(kgrid, axis=-1)

    knorm = (kgrid**2).sum(axis=-1) + np.finfo(np.float32).eps
    dipole_kernel = 1 / 3 - (kgrid[..., 0] ** 2 / knorm)

    # apply convolution
    B0map = ifftc(
        dipole_kernel * fftc(chi, ax=range(-ndim, 0)), ax=range(-ndim, 0)
    ).real

    # rescale
    if b0range is not None:
        B0map = B0map - B0map.min()  # (min, max) -> (0, max - min)
        B0map = B0map / B0map.max()  # (0, max - min) -> (0, 1)
        B0map = (
            B0map * (b0range[1] - b0range[0]) + b0range[0]
        )  # (0, 1) -> (b0range[0], b0range[1])
    else:
        gamma = 42.58 * 1e6  # Hz / T
        scale = gamma * B0
        B0map = scale * B0map

    # mask
    if mask is not None:
        mask = mask != 0
        B0map = mask * B0map

    return B0map.astype(np.float32)
